#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
from cfnbootstrap import security, util
from cfnbootstrap.construction_errors import ToolError
from contextlib import contextmanager
import base64
import logging
import os
from cfnbootstrap.packages import requests
import stat
import codecs

_templates_supported = True
try:
    from pystache.renderer import Renderer
except ImportError:
    _templates_supported = False

try:
    import simplejson as json
except ImportError:
    import json

log = logging.getLogger("cfn.init")


class FileTool(object):
    """
    Writes files to disk

    """

    _compare_buffer = 8 * 1024

    @classmethod
    def is_same_file(cls, f1, f2):
        if hasattr(os.path, "samefile"):
            return os.path.samefile(f1, f2)
        else:
            #Crude workaround for os.path.samefile only existing on Unix
            return os.path.normcase(os.path.abspath(f1)) == os.path.normcase(os.path.abspath(f2))

    @classmethod
    def compare_file_contents(cls, f1, f2):
        """
        Return true if f1 and f2 have the same content.
        """
        if os.path.getsize(f1) != os.path.getsize(f2):
            return False

        if cls.is_same_file(f1, f2):
            return True

        # Borrowed from filecmp
        with file(f1, 'rb') as fp1:
            with file(f2, 'rb') as fp2:
                bufsize = 8 * 1024
                while True:
                    b1 = fp1.read(bufsize)
                    b2 = fp2.read(bufsize)
                    if b1 != b2:
                        return False
                    if not b1:
                        return True

    def apply(self, action, auth_config):
        """
        Write a set of files to disk, returning a list of the files that have changed.

        Arguments:
        action -- a dict of pathname to attributes, such as owner, group, mode, content, and encoding
        auth_config -- an AuthenticationConfig object for managing authenticated downloads

        Exceptions:
        ToolError -- on expected failures
        """

        files_changed = []

        if not action.keys():
            log.debug("No files specified")
            return files_changed

        for (filename, attribs) in sorted(action.iteritems(), key=lambda pair: pair[0]):
            if not os.path.isabs(filename):
                raise ToolError('File specified with non-absolute path: %s' % filename)


            # The only difference between a file and a symlink is hidden in the mode
            file_is_link = "mode" in attribs and stat.S_ISLNK(int(attribs["mode"], 8))

            if file_is_link:
                if "content" not in attribs:
                    raise ToolError("Symbolic link specified without a destination")
                elif os.path.exists(filename) and FileTool.is_same_file(os.path.realpath(filename), attribs["content"]):
                    log.info("Symbolic link %s already exists", filename)
                    continue

            parent = os.path.dirname(filename)
            if not os.path.isdir(parent):
                if not os.path.exists(parent):
                    log.debug("Parent directory %s does not exist, creating", parent)
                    os.makedirs(parent)
                else:
                    raise ToolError("Parent directory %s exists and is a file" % parent)

            currentFileExists = False
            # if file exists, copy its current mode
            if os.path.isfile(filename):
                currentFileExists = True
                currentFileMode = stat.S_IMODE(os.stat(filename).st_mode)

            with self.backup(filename, files_changed):
                if file_is_link:
                    log.debug("%s is specified as a symbolic link to %s", filename, attribs['content'])
                    os.symlink(attribs["content"], filename)
                else:
                    file_is_text = 'content' in attribs and not self._is_base64(attribs)
                    log.debug("Writing content to %s", filename)
                    if file_is_text:
                        with codecs.open(filename, 'w', encoding='utf-8') as f:
                            self._write_file(f, attribs, auth_config)
                    else:
                        with file(filename, 'wb') as f:
                            self._write_file(f, attribs, auth_config)

                if "mode" in attribs:
                    log.debug("Setting mode for %s to %s", filename, attribs["mode"])
                    os.chmod(filename, stat.S_IMODE(int(attribs["mode"], 8)))
                else:
                    if currentFileExists:
                        log.debug("No mode specified for %s. Setting permissions of the existing file: %s.", filename, oct(currentFileMode))
                        os.chmod(filename, currentFileMode)
                    else:
                        # get attributes that the file was just created
                        currentFileMode = stat.S_IMODE(os.stat(filename).st_mode)
                        # filter the curent mode by giving up others writable
                        currentFileMode = currentFileMode & 0775
                        log.debug("No mode specified for %s. The file will be created with the mode: %s", filename, oct(currentFileMode))
                        os.chmod(filename, currentFileMode)

                security.set_owner_and_group(filename, attribs.get("owner"), attribs.get("group"))

        return files_changed

    @contextmanager
    def backup(self, filename, files_changed):
        backup_file = None
        backup_backup_file = None
        if os.path.exists(filename):
            log.debug("%s already exists", filename)
            backup_file = filename + '.bak'
            if os.path.exists(backup_file):
                backup_backup_file = backup_file + "2"
                if os.path.exists(backup_backup_file):
                    os.remove(backup_backup_file)
                self._backup_file(backup_file, backup_backup_file)
            self._backup_file(filename, backup_file)

        try:
            yield backup_file
        except Exception:
            if backup_file:
                try:
                    if os.path.exists(filename):
                        os.remove(filename)
                    self._backup_file(backup_file, filename)
                    if backup_backup_file:
                        self._backup_file(backup_backup_file, backup_file)
                except ToolError, t:
                    log.warn("Error restoring %s from backup", filename)
            raise
        else:
            linkmode = backup_file and os.path.islink(backup_file) or os.path.islink(filename)
            # we assume any symbolic links changed because we short-circuit links to the same files early on
            if not backup_file or linkmode or not FileTool.compare_file_contents(backup_file, filename):
                files_changed.append(filename)
                if backup_backup_file:
                    os.remove(backup_backup_file)
            elif backup_file and backup_backup_file:
                try:
                    self._backup_file(backup_backup_file, backup_file)
                except ToolError, t:
                    log.warn("Error restoring backup file %s: %s", backup_file, str(t))

    def _backup_file(self, source, dest):
        try:
            log.debug("Moving %s to %s", source, dest)
            os.rename(source, dest)
        except OSError, e:
            log.error("Could not move %s to %s", source, dest)
            raise ToolError("Could not rename %s: %s" % (source, str(e)))

    def _is_base64(self, attribs):
        return attribs.get("encoding", "plain") == "base64"

    def _write_file(self, dest_fileobj, attribs, auth_config):
        content = attribs.get("content", "")
        if content:
            self._write_inline_content(dest_fileobj, content, self._is_base64(attribs),
                                       attribs.get('context'))
        else:
            source = attribs.get("source", "")
            if not source:
                raise ToolError("File specified without source or content")
            log.debug("Retrieving contents from %s", source)

            try:
                self._write_remote_file(source, auth_config.get_auth(attribs.get('authentication', None)), dest_fileobj,
                                      attribs.get('context'))
            except IOError, e:
                raise ToolError("Failed to retrieve %s: %s" % (source, e.strerror))

    def _render_template(self, content, context):
        if not _templates_supported:
            raise ToolError("Pystache must be installed in order to render files as Mustache templates")

        log.debug('Rendering as Mustache template')
        try:
            return Renderer(string_encoding='utf-8', file_encoding='utf-8').render(content, context)
        except Exception, e:
            raise ToolError("Failed to render content as a Mustache template: %s" % e.message)

    @util.retry_on_failure()
    def _write_remote_file(self, source, auth, dest, context):
        opts = util.req_opts({'auth': auth})
        remote_contents = util.EtagCheckedResponse(requests.get(source, **opts))

        if context is None:
            remote_contents.write_to(dest)
        else:
            dest.write(self._render_template(remote_contents.contents(), context))

    def _write_inline_content(self, dest, content, is_base64, context):
        if not isinstance(content, basestring):
            log.debug('Content will be serialized as a JSON structure')
            json.dump(content, dest)
            return

        if is_base64:
            try:
                log.debug("Decoding base64 content")
                dest.write(base64.b64decode(content.strip()))
            except TypeError:
                raise ToolError("Malformed base64: %s" % content)
        elif context is None:
            dest.write(content)
        else:
            dest.write(self._render_template(content, context))



#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
import endpoint_tool
import hashlib
from optparse import OptionGroup
import threading
import datetime
from cfnbootstrap.packages.requests.exceptions import ConnectionError, HTTPError, Timeout, SSLError
import StringIO
import imp
import logging
import os.path
import random
import re
from cfnbootstrap.packages import requests
import stat
import subprocess
import sys
import time

try:
    import simplejson as json
except ImportError:
    import json

try:
    from cfnbootstrap.packages.requests.packages.urllib3.exceptions import ProxyError
except ImportError:
    class ProxyError(Exception):
        def __init__(self, *args, **kwargs):
            super(ProxyError, self).__init__(*args, **kwargs)

log = logging.getLogger("cfn.init")
wire_log = logging.getLogger("wire")
cmd_log = logging.getLogger("cfn.init.cmd")

#==============================================================================
# HTTP backoff-and-retry
#==============================================================================

def exponential_backoff(max_tries, max_sleep=20):
    """
    Returns a series of floating point numbers between 0 and min(max_sleep, 2^i-1) for i in 0 to max_tries
    """
    return [random.random() * min(max_sleep, (2**i - 1)) for i in range(0, max_tries)]

def extend_backoff(durations, max_sleep=20):
    """
    Adds another exponential delay time to a list of delay times
    """
    durations.append(random.random() * min(max_sleep, (2**len(durations) - 1)))

def _extract_http_error(resp):
    if resp.status_code == 503:
        retry_mode='RETRIABLE_FOREVER'
    elif resp.status_code < 500 and resp.status_code not in (404, 408):
        retry_mode='TERMINAL'
    else:
        retry_mode='RETRIABLE'

    return RemoteError(resp.status_code, "HTTP Error %s : %s" % (resp.status_code, resp.text), retry_mode)

class EtagCheckedResponse(object):

    def __init__(self, response):
        self._response = check_status(response)
        etag = response.headers['etag'].strip('"') if 'etag' in response.headers and endpoint_tool.is_service_url("AmazonS3", response.url) else None
        if response.headers.get('x-amz-server-side-encryption') == 'aws:kms':
            log.warn('file uses KMS encryption; skipping checksum comparison')
            etag = None
        elif etag and '-' in etag:
            log.warn('cannot check consistency of file uploaded multipart; etag has - character present')
            etag = None

        self._etag = etag
        self._digest = hashlib.md5() if self._etag else NoOpDigest()

    def _check_digest(self):
        if not self._etag:
            return

        final_digest = self._digest.hexdigest()
        if self._etag != final_digest:
            raise ChecksumError("Expected digest %s; received %s" % (self._etag, final_digest))

    def write_to(self, dest):
        dest.seek(0, 0)
        dest.truncate()
        for c in self._response.iter_content(10 * 1024):
            dest.write(c)
            self._digest.update(c)
        self._check_digest()

    def contents(self):
        content = self._response.content
        self._digest.update(content)
        self._check_digest()
        return content

class ChecksumError(IOError):

    def __init__(self, msg):
        super(ChecksumError, self).__init__(None, msg)

class NoOpDigest():

    def __init__(self):
        self.digest_size = -1
        self.block_size = -1

    def update(self, content):
        pass

    def hexdigest(self):
        return None

    def digest(self):
        return None

    def copy(self):
        return self

class RemoteError(IOError):

    retry_modes = frozenset(['TERMINAL', 'RETRIABLE', 'RETRIABLE_FOREVER'])

    def __init__(self, code, msg, retry_mode='RETRIABLE'):
        super(RemoteError, self).__init__(code, msg)
        if not retry_mode in RemoteError.retry_modes:
            raise ValueError("Invalid retry mode: %s" % retry_mode)
        self.retry_mode = retry_mode

class Sleeper(object):

    def sleep(self, secs):
        time.sleep(secs)

def retry_on_failure(max_tries = 5, http_error_extractor=_extract_http_error):
    def _decorate(f):
        def _retry(*args, **kwargs):
            sleeper = Sleeper()
            durations = exponential_backoff(max_tries)
            for duration in durations:
                if duration > 0:
                    log.debug("Sleeping for %f seconds before retrying", duration)
                    sleeper.sleep(duration)

                try:
                    return f(*args, **kwargs)
                except SSLError, e:
                    log.exception("SSLError")
                    last_error = RemoteError(None, str(e))
                except ChecksumError, e:
                    log.exception("Checksum mismatch")
                    last_error = RemoteError(None, str(e))
                except ConnectionError, e:
                    log.exception('ConnectionError')
                    last_error = RemoteError(None, str(e))
                except ProxyError, e:
                    log.exception('ProxyError')
                    last_error = RemoteError(None, str(e))
                    # ProxyError skips the typical 3 retries done by urllib
                    # this prevents us from taking an availability hit when there is a 'false' ProxyError
                    # which happens because Requests never passes proxies==None to urllib3,
                    # and newer versions of urllib3 always wrap socket errors in ProxyError when proxies is not None
                    if len(durations) < max_tries * 3:
                        extend_backoff(durations)
                except HTTPError, e:
                    last_error = http_error_extractor(e.response)
                    if last_error.retry_mode == 'TERMINAL':
                        raise last_error
                    elif last_error.retry_mode == 'RETRIABLE_FOREVER':
                        extend_backoff(durations)

                    log.exception(last_error.strerror)
                except Timeout, e:
                    log.exception('Timeout')
                    last_error = RemoteError(None, str(e))
                except TimeoutError, e:
                    log.exception('Client-side timeout')
                    last_error = RemoteError(None, str(e))
                except IOError, e:
                    log.exception('Generic IOError')
                    last_error = RemoteError(None, str(e))
                except Exception, e:
                    log.exception('Unexpected Exception')
                    raise RemoteError(None, str(e), 'TERMINAL')
            else:
                raise last_error
        return _retry
    return _decorate

class TimeoutError(StandardError):

    def __init__(self, msg):
        super(StandardError, self).__init__()
        self.msg = msg


def timeout(duration=60):
    def _decorate(f):
        def _timeout(*args, **kwargs):
            ret_val = []
            exc = []
            def value_fn():
                try:
                    ret_val.append(f(*args, **kwargs))
                except Exception, e:
                    exc.append(e)

            worker_thread = threading.Thread(target=value_fn)
            worker_thread.daemon = True
            worker_thread.start()
            worker_thread.join(duration)

            if worker_thread.isAlive():
                log.warn('Timeout of %s seconds breached', duration)
                raise TimeoutError("Execution did not succeed after %s seconds" % duration)

            if exc:
                raise exc[0]

            return ret_val[0] if ret_val else None

        return _timeout
    return _decorate


#==============================================================================
# Instance metadata
#==============================================================================

@retry_on_failure(max_tries=10)
def get_instance_identity_document():
    resp = requests.get('http://169.254.169.254/latest/dynamic/instance-identity/document', proxies = {'no_proxy' : '169.254.169.254/32'})
    resp.raise_for_status()
    return resp.text.rstrip()

@retry_on_failure(max_tries=10)
def get_instance_identity_signature():
    resp = requests.get('http://169.254.169.254/latest/dynamic/instance-identity/signature', proxies = {'no_proxy' : '169.254.169.254/32'})
    resp.raise_for_status()
    return resp.text.rstrip()

_instance_id = '__unset'

@retry_on_failure(max_tries=10)
def _fetch_instance_id():
    resp = requests.get('http://169.254.169.254/latest/meta-data/instance-id', timeout=2, proxies = {'no_proxy' : '169.254.169.254/32'})
    resp.raise_for_status()
    return resp.text.strip()

def get_instance_id():
    """
    Attempt to retrieve an EC2 instance id, returning None if this is not EC2
    """
    global _instance_id
    if _instance_id == '__unset':
        try:
            _instance_id = _fetch_instance_id()
        except IOError:
            log.exception("Exception retrieving InstanceId")
            _instance_id =  None

    return _instance_id

def is_ec2():
    return get_instance_id() is not None

@retry_on_failure(max_tries=10)
def get_role_creds(name):
    resp = requests.get('http://169.254.169.254/latest/meta-data/iam/security-credentials/%s' % name, proxies = {'no_proxy' : '169.254.169.254/32'})
    resp.raise_for_status()
    role = resp.json()
    return Credentials(role['AccessKeyId'], role['SecretAccessKey'], role['Token'], datetime.datetime.strptime(role['Expiration'], '%Y-%m-%dT%H:%M:%SZ'))

_trues = frozenset([True, 1, 'true', 'yes', 'y', '1'])

#==============================================================================
# Miscellaneous
#==============================================================================

def interpret_boolean(input):
    """
    This tries to interpret if the user intended True
    I don't use python's boolean equivalent because it's
    likely that we're getting a string
    """
    if not input:
        return False

    input = input.lower().strip() if isinstance(input, basestring) else input

    return input in _trues

_dot_split = re.compile(r'(?<!\\)\.')
_slash_replace = re.compile(r'\\(?=\.)')

def extract_value(metadata, path):
    """Returns a value from metadata (a dict) at a (possibly empty) path, where path is in dotted object syntax (like root.child.leaf)"""
    if not path:
        return metadata

    return_data = metadata
    for element in (_slash_replace.sub('', s) for s in _dot_split.split(path)):
        if not element in return_data:
            log.debug("No value at path %s (missing index: %s)", path, element)
            return None
        return_data = return_data[element]

    return return_data

#==============================================================================
# Command-line (credentials, options, etc)
#==============================================================================

def get_proxy_options(parser):
    proxy_group = OptionGroup(parser, "Proxy", "Options for specifying proxies. Format: [scheme://][user:password@]host:port")

    proxy_group.add_option("", "--http-proxy", help="A (non-SSL) HTTP proxy", type="string", dest="http_proxy")
    proxy_group.add_option("", "--https-proxy", help="An HTTPS proxy", type="string", dest="https_proxy")

    return proxy_group

def get_proxyinfo(options):
    return_info = {}
    if options.http_proxy:
        return_info['http'] = options.http_proxy
    if options.https_proxy:
        return_info['https'] = options.https_proxy

    return return_info if return_info else None

def get_cred_options(parser):
    creds_group = OptionGroup(parser, "AWS Credentials", "Options for specifying AWS Account Credentials.")

    creds_group.add_option("-f", "--credential-file", help="A credential file, readable only by the owner, with keys "
                                                           "'AWSAccessKeyId' and 'AWSSecretKey'",
                           type="string", dest="credential_file")

    creds_group.add_option("", "--role", help="An IAM Role",
        type="string", dest="iam_role")

    creds_group.add_option("", "--access-key", help="An AWS Access Key",
                           type="string", dest="access_key")
    creds_group.add_option("", "--secret-key", help="An AWS Secret Key",
                           type="string", dest="secret_key")

    return creds_group

def get_creds_or_die(options):
    if options.credential_file:
        try:
            return extract_credentials(options.credential_file)
        except IOError, e:
            print >> sys.stderr, "Error retrieving credentials from file:\n\t%s" % e.strerror
            sys.exit(1)
    elif options.iam_role:
        return get_role_creds(options.iam_role)
    else:
        return Credentials(options.access_key, options.secret_key)


def extract_credentials(path):
    """
    Extract credentials from a file at path, returning tuple of (access_key, secret_key)
    Raises an exception if the file is readable by group or other.
    """
    if not os.path.isfile(path):
        raise IOError(None, "Credential file was not found at %s" % path)

    if os.name == 'posix':
        mode = os.stat(path)[stat.ST_MODE]

        if stat.S_IRWXG & mode or stat.S_IRWXO & mode:
            raise IOError(None, "Credential file cannot be accessible by group or other. Please chmod 600 the credential file.")

    access_key, secret_key = '', ''
    with file(path, 'r') as f:
        for line in (line.strip() for line in f):
            if line.startswith("AWSAccessKeyId="):
                access_key = line.partition('=')[2]
            elif line.startswith("AWSSecretKey="):
                secret_key = line.partition('=')[2]

    if not access_key or not secret_key:
        raise IOError(None, "Credential file must contain the keys 'AWSAccessKeyId' and 'AWSSecretKey'")

    return Credentials(access_key, secret_key)

#==============================================================================
# Process running utilities
#==============================================================================

class ProcessResult(object):
    """
    Return object for ProcessHelper

    """

    def __init__(self, returncode, stdout, stderr):
        self._returncode = returncode
        self._stdout = stdout
        self._stderr = stderr

    @property
    def returncode(self):
        return self._returncode

    @property
    def stdout(self):
        return self._stdout

    @property
    def stderr(self):
        return self._stderr

class ProcessHelper(object):
    """
    Helper to simplify command line execution

    """

    def __init__(self, cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=None, cwd=None):
        self._cmd = cmd
        self._stdout = stdout
        self._stderr = stderr
        if not env:
            self._env = None
        elif os.name == 'nt': # stringify the environment in Windows, which cannot handle unicodes
            self._env = dict(((str(k), str(v)) for k,v in env.iteritems()))
        else:
            self._env = dict(env)
        self._cwd = cwd

    def call(self):
        """
        Calls the command, returning a tuple of (returncode, stdout, stderr)
        """

        process = subprocess.Popen(self._cmd, stdout=self._stdout, stderr=self._stderr,
                                   shell=isinstance(self._cmd, basestring), env=self._env, cwd=self._cwd)
        returnData = process.communicate()

        return ProcessResult(process.returncode, returnData[0], returnData[1])

class LoggingProcessHelper(object):

    def __init__(self, cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, name='', env=None, cwd=None):
        self.name = name
        self.process_helper = ProcessHelper(cmd,stdout,stderr,env,cwd)
        self.cmd = cmd

    def call(self):
        cmd_log.info(60*"=")
        cmd_log.info(self.name)
        results = self.process_helper.call()
        output = results.stdout
        stderr = results.stderr
        if output:
            cmd_log.info("Command Output".center(60,'-'))
            for line in output.splitlines(False):
                cmd_log.info('\t' + line)
            cmd_log.info(60*"-")
        if stderr:
            cmd_log.info("Command Errors".center(60,'-'))
            for line in stderr.splitlines(False):
                cmd_log.error('\t' + line)
            cmd_log.info(60*"-")
        if results.returncode:
            cmd_log.error("Exited with error code %d", results.returncode)
        else:
            cmd_log.info("Completed successfully.")
        return results



class Credentials(object):
    '''
    AWS Credentials
    '''

    def __init__(self, access_key, secret_key, security_token=None, expiration=None):
        self._access_key = access_key
        self._secret_key = secret_key
        self._security_token = security_token
        self._expiration = expiration

    @property
    def access_key(self):
        return self._access_key

    @property
    def secret_key(self):
        return self._secret_key

    @property
    def security_token(self):
        return self._security_token

    @property
    def expiration(self):
        return self._expiration


class RoleBasedCredentials(object):
    '''
    Refreshing credentials
    '''

    def __init__(self, role_name):
        self._role_name = role_name
        self._creds = None
        self._refresh_lock = threading.Lock()
        self._refresh()

    def _refresh(self):
        with self._refresh_lock:
            if not self._creds or self._creds.expiration - datetime.timedelta(hours=2) < datetime.datetime.utcnow():
                log.info('Refreshing role-based credentials')
                try:
                    self._creds = get_role_creds(self._role_name)
                except IOError:
                    if not self._creds or self._creds.expiration < datetime.datetime.utcnow():
                        raise RuntimeError('Credentials have expired and refresh failed.')
                    log.exception('Exception while refreshing credentials; will refresh on next access')

            return self._creds

    @property
    def access_key(self):
        return self._refresh().access_key

    @property
    def secret_key(self):
        return self._refresh().secret_key

    @property
    def security_token(self):
        return self._refresh().security_token

    @property
    def expiration(self):
        return self._refresh().expiration

def log_response(resp, *args, **kwargs):
    wire_log.debug('Response: %s %s [headers: %s]', resp.status_code, resp.url, resp.headers)
    if not resp.ok:
        wire_log.debug('Response error: %s', resp.content)

def get_hooks():
    return {'response': log_response}

def req_opts(kwargs):
    kwargs = dict(kwargs) if kwargs else {}
    kwargs['verify'] = True
    kwargs['hooks'] = get_hooks()
    kwargs['stream'] = True

    return kwargs

def check_status(resp):
    resp.raise_for_status()
    return resp



import datetime
import sys
import socket
from socket import timeout as SocketTimeout
import warnings
from .packages import six

try:  # Python 3
    from http.client import HTTPConnection as _HTTPConnection, HTTPException
except ImportError:
    from httplib import HTTPConnection as _HTTPConnection, HTTPException


class DummyConnection(object):
    "Used to detect a failed ConnectionCls import."
    pass


try:  # Compiled with SSL?
    HTTPSConnection = DummyConnection
    import ssl
    BaseSSLError = ssl.SSLError
except (ImportError, AttributeError):  # Platform-specific: No SSL.
    ssl = None

    class BaseSSLError(BaseException):
        pass


try:  # Python 3:
    # Not a no-op, we're adding this to the namespace so it can be imported.
    ConnectionError = ConnectionError
except NameError:  # Python 2:
    class ConnectionError(Exception):
        pass


from .exceptions import (
    ConnectTimeoutError,
    SystemTimeWarning,
    SecurityWarning,
)
from .packages.ssl_match_hostname import match_hostname

from .util.ssl_ import (
    resolve_cert_reqs,
    resolve_ssl_version,
    ssl_wrap_socket,
    assert_fingerprint,
)


from .util import connection

port_by_scheme = {
    'http': 80,
    'https': 443,
}

RECENT_DATE = datetime.date(2014, 1, 1)


class HTTPConnection(_HTTPConnection, object):
    """
    Based on httplib.HTTPConnection but provides an extra constructor
    backwards-compatibility layer between older and newer Pythons.

    Additional keyword parameters are used to configure attributes of the connection.
    Accepted parameters include:

      - ``strict``: See the documentation on :class:`urllib3.connectionpool.HTTPConnectionPool`
      - ``source_address``: Set the source address for the current connection.

        .. note:: This is ignored for Python 2.6. It is only applied for 2.7 and 3.x

      - ``socket_options``: Set specific options on the underlying socket. If not specified, then
        defaults are loaded from ``HTTPConnection.default_socket_options`` which includes disabling
        Nagle's algorithm (sets TCP_NODELAY to 1) unless the connection is behind a proxy.

        For example, if you wish to enable TCP Keep Alive in addition to the defaults,
        you might pass::

            HTTPConnection.default_socket_options + [
                (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
            ]

        Or you may want to disable the defaults by passing an empty list (e.g., ``[]``).
    """

    default_port = port_by_scheme['http']

    #: Disable Nagle's algorithm by default.
    #: ``[(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)]``
    default_socket_options = [(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)]

    #: Whether this connection verifies the host's certificate.
    is_verified = False

    def __init__(self, *args, **kw):
        if six.PY3:  # Python 3
            kw.pop('strict', None)

        # Pre-set source_address in case we have an older Python like 2.6.
        self.source_address = kw.get('source_address')

        if sys.version_info < (2, 7):  # Python 2.6
            # _HTTPConnection on Python 2.6 will balk at this keyword arg, but
            # not newer versions. We can still use it when creating a
            # connection though, so we pop it *after* we have saved it as
            # self.source_address.
            kw.pop('source_address', None)

        #: The socket options provided by the user. If no options are
        #: provided, we use the default options.
        self.socket_options = kw.pop('socket_options', self.default_socket_options)

        # Superclass also sets self.source_address in Python 2.7+.
        _HTTPConnection.__init__(self, *args, **kw)

    def _new_conn(self):
        """ Establish a socket connection and set nodelay settings on it.

        :return: New socket connection.
        """
        extra_kw = {}
        if self.source_address:
            extra_kw['source_address'] = self.source_address

        if self.socket_options:
            extra_kw['socket_options'] = self.socket_options

        try:
            conn = connection.create_connection(
                (self.host, self.port), self.timeout, **extra_kw)

        except SocketTimeout:
            raise ConnectTimeoutError(
                self, "Connection to %s timed out. (connect timeout=%s)" %
                (self.host, self.timeout))

        return conn

    def _prepare_conn(self, conn):
        self.sock = conn
        # the _tunnel_host attribute was added in python 2.6.3 (via
        # http://hg.python.org/cpython/rev/0f57b30a152f) so pythons 2.6(0-2) do
        # not have them.
        if getattr(self, '_tunnel_host', None):
            # TODO: Fix tunnel so it doesn't depend on self.sock state.
            self._tunnel()
            # Mark this connection as not reusable
            self.auto_open = 0

    def connect(self):
        conn = self._new_conn()
        self._prepare_conn(conn)


class HTTPSConnection(HTTPConnection):
    default_port = port_by_scheme['https']

    def __init__(self, host, port=None, key_file=None, cert_file=None,
                 strict=None, timeout=socket._GLOBAL_DEFAULT_TIMEOUT, **kw):

        HTTPConnection.__init__(self, host, port, strict=strict,
                                timeout=timeout, **kw)

        self.key_file = key_file
        self.cert_file = cert_file

        # Required property for Google AppEngine 1.9.0 which otherwise causes
        # HTTPS requests to go out as HTTP. (See Issue #356)
        self._protocol = 'https'

    def connect(self):
        conn = self._new_conn()
        self._prepare_conn(conn)
        self.sock = ssl.wrap_socket(conn, self.key_file, self.cert_file)


class VerifiedHTTPSConnection(HTTPSConnection):
    """
    Based on httplib.HTTPSConnection but wraps the socket with
    SSL certification.
    """
    cert_reqs = None
    ca_certs = None
    ssl_version = None
    assert_fingerprint = None

    def set_cert(self, key_file=None, cert_file=None,
                 cert_reqs=None, ca_certs=None,
                 assert_hostname=None, assert_fingerprint=None):

        self.key_file = key_file
        self.cert_file = cert_file
        self.cert_reqs = cert_reqs
        self.ca_certs = ca_certs
        self.assert_hostname = assert_hostname
        self.assert_fingerprint = assert_fingerprint

    def connect(self):
        # Add certificate verification
        conn = self._new_conn()

        resolved_cert_reqs = resolve_cert_reqs(self.cert_reqs)
        resolved_ssl_version = resolve_ssl_version(self.ssl_version)

        hostname = self.host
        if getattr(self, '_tunnel_host', None):
            # _tunnel_host was added in Python 2.6.3
            # (See: http://hg.python.org/cpython/rev/0f57b30a152f)

            self.sock = conn
            # Calls self._set_hostport(), so self.host is
            # self._tunnel_host below.
            self._tunnel()
            # Mark this connection as not reusable
            self.auto_open = 0

            # Override the host with the one we're requesting data from.
            hostname = self._tunnel_host

        is_time_off = datetime.date.today() < RECENT_DATE
        if is_time_off:
            warnings.warn((
                'System time is way off (before {0}). This will probably '
                'lead to SSL verification errors').format(RECENT_DATE),
                SystemTimeWarning
            )

        # Wrap socket using verification with the root certs in
        # trusted_root_certs
        self.sock = ssl_wrap_socket(conn, self.key_file, self.cert_file,
                                    cert_reqs=resolved_cert_reqs,
                                    ca_certs=self.ca_certs,
                                    server_hostname=hostname,
                                    ssl_version=resolved_ssl_version)

        if self.assert_fingerprint:
            assert_fingerprint(self.sock.getpeercert(binary_form=True),
                               self.assert_fingerprint)
        elif resolved_cert_reqs != ssl.CERT_NONE \
                and self.assert_hostname is not False:
            cert = self.sock.getpeercert()
            if not cert.get('subjectAltName', ()):
                warnings.warn((
                    'Certificate has no `subjectAltName`, falling back to check for a `commonName` for now. '
                    'This feature is being removed by major browsers and deprecated by RFC 2818. '
                    '(See https://github.com/shazow/urllib3/issues/497 for details.)'),
                    SecurityWarning
                )
            match_hostname(cert, self.assert_hostname or hostname)

        self.is_verified = (resolved_cert_reqs == ssl.CERT_REQUIRED
                            or self.assert_fingerprint is not None)


if ssl:
    # Make a copy for testing.
    UnverifiedHTTPSConnection = HTTPSConnection
    HTTPSConnection = VerifiedHTTPSConnection



#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
from cfnbootstrap.construction_errors import ToolError
from cfnbootstrap.util import ProcessHelper, LoggingProcessHelper
import logging
import re
import subprocess

log = logging.getLogger("cfn.init")

class YumTool(object):
    """
    Installs packages via Yum

    """

    def apply(self, action, auth_config=None):
        """
        Install a set of packages via yum, returning the packages actually installed or updated.

        Arguments:
        action -- a dict of package name to version; version can be empty, a single string or a list of strings

        Exceptions:
        ToolError -- on expected failures (such as a non-zero exit code)
        """

        pkgs_changed = []

        if not action.keys():
            log.debug("No packages specified for yum")
            return pkgs_changed

        cache_result = ProcessHelper(['yum', '-y', 'makecache']).call()

        if cache_result.returncode:
            log.error("Yum makecache failed. Output: %s", cache_result.stdout)
            raise ToolError("Could not create yum cache", cache_result.returncode)

        pkg_specs_to_upgrade = []
        pkg_specs_to_downgrade = []

        for pkg_name in action:
            if action[pkg_name]:
                if isinstance(action[pkg_name], basestring):
                    pkg_ver = action[pkg_name]
                else:
                    # Yum only cares about one version anyway... so take the max specified version in the list
                    pkg_ver = RpmTool.max_version(action[pkg_name])
            else:
                pkg_ver = None

            pkg_spec = '%s-%s' % (pkg_name, pkg_ver) if pkg_ver else pkg_name

            if self._pkg_installed(pkg_spec):
                # If the EXACT requested spec is installed, don't do anything
                log.debug("%s will not be installed as it is already present", pkg_spec)
            elif not self._pkg_available(pkg_spec):
                # If the requested spec is not available, blow up
                log.error("%s is not available to be installed", pkg_spec)
                raise ToolError("Yum does not have %s available for installation" % pkg_spec)
            elif not pkg_ver:
                # If they didn't request a specific version, always upgrade
                pkg_specs_to_upgrade.append(pkg_spec)
                pkgs_changed.append(pkg_name)
            else:
                # They've requested a specific version that's available but not installed.
                # Figure out if it's an upgrade or a downgrade
                installed_version = RpmTool.get_package_version(pkg_name, False)[1]
                if self._should_upgrade(pkg_ver, installed_version):
                    pkg_specs_to_upgrade.append(pkg_spec)
                    pkgs_changed.append(pkg_name)
                else:
                    log.debug("Downgrading to %s from installed version %s", pkg_spec, installed_version)
                    pkg_specs_to_downgrade.append(pkg_spec)
                    pkgs_changed.append(pkg_name)


        if not pkgs_changed:
            log.debug("All yum packages were already installed")
            return []

        if pkg_specs_to_upgrade:
            log.debug("Installing/updating %s via yum", pkg_specs_to_upgrade)

            result = LoggingProcessHelper(['yum', '-y', 'install'] + pkg_specs_to_upgrade, name=u'yum install %s' % ' '.join(pkg_specs_to_upgrade)).call()

            if result.returncode:
                log.error("Yum failed. Output: %s", result.stdout)
                raise ToolError("Could not successfully install/update yum packages", result.returncode)

        if pkg_specs_to_downgrade:
            log.debug("Downgrading %s via yum", pkg_specs_to_downgrade)

            result = LoggingProcessHelper(['yum', '-y', 'downgrade'] + pkg_specs_to_downgrade, name=u'yum downgrade %s' % ' '.join(pkg_specs_to_downgrade)).call()

            if result.returncode:
                log.error("Yum failed. Output: %s", result.stdout)
                raise ToolError("Could not successfully downgrade yum packages", result.returncode)


        log.info("Yum installed %s", pkgs_changed)

        return pkgs_changed


    def _should_upgrade(self, requested_ver, installed_version):
        # If they haven't requested a version, always install
        if not requested_ver:
            return True
        #Now we need to detect whether or not we need to upgrade
        ver_cmp = RpmTool.compare_rpm_versions(requested_ver, installed_version)
        if ver_cmp > 0:
            log.debug("Requested version %s is greater than installed version %s, so we will upgrade", requested_ver, installed_version)
            return True
        else:
            log.debug("Requested version %s is NOT greater than installed version %s, so we will NOT upgrade", requested_ver, installed_version)
            return False

    def _pkg_installed(self, pkg):
        result = ProcessHelper(['yum', '-C', '-y', 'list', 'installed', pkg]).call()

        return result.returncode == 0

    def _pkg_available(self, pkg):
        # --showduplicates seems to be required to see downgradable versions when running yum non-interactively
        # but not when running interactively -- but we rarely run interactively
        result = ProcessHelper(['yum', '-C', '-y', '--showduplicates', 'list', 'available', pkg]).call()

        return result.returncode == 0

class RpmTool(object):

    def apply(self, action, auth_config=None):
        """
        Install a set of packages via RPM, returning the packages actually installed or updated.

        Arguments:
        action -- a dict of package name to version; version can be empty, a single string or a list of strings

        Exceptions:
        ToolError -- on expected failures (such as a non-zero exit code)
        """

        pkgs_changed = []

        if not action.keys():
            log.debug("No packages installed for RPM")
            return pkgs_changed

        pkgs = []

        for pkg_name, loc in action.iteritems():
            pkgs_to_process = ([loc] if isinstance(loc, basestring) else loc)
            pkgs_filtered = [pkg_key for pkg_key in pkgs_to_process if self._package_filter(pkg_key)]
            if pkgs_filtered:
                pkgs.extend(pkgs_filtered)
                pkgs_changed.append(pkg_name)


        if not pkgs:
            log.info("All RPMs were already installed")
            return []

        log.debug("Installing %s via RPM", pkgs)

        result = ProcessHelper(['rpm', '-U', '--quiet', '--nosignature', '--replacepkgs'] + pkgs).call()

        if result.returncode:
            log.error("RPM failed. Output: %s", result.stdout)
            raise ToolError("Could not successfully install rpm packages", result.returncode)
        else:
            log.debug("RPM output: %s", result.stdout)

        return pkgs_changed

    def _package_filter(self, pkg):
        if not pkg:
            log.warn("RPM specified with no location")
            return False

        if self._is_installed(pkg):
            log.debug("Skipping RPM at %s as it is already installed", pkg)
            return False

        return True

    @classmethod
    def get_package_version(cls, pkg, is_file=True):
        """
        Given the name of an installed package or package location, return a tuple of (name, version-release)
        of either the installed package or the specified package location

        Parameters:
            - pkg: the package name/location
            - is_file : if True, pkg refers to a package location; if False, the name of an installed package
        """

        query_mode = '-qp' if is_file else '-qa'

        log.debug("Querying for version of package %s", pkg)

        query_result = ProcessHelper(['rpm', query_mode, '--queryformat', '%{NAME}|%{VERSION}-%{RELEASE}', '--nosignature', pkg], stderr=subprocess.PIPE).call()

        log.debug("RPM stdout: %s", query_result.stdout)
        log.debug("RPM stderr: %s", query_result.stderr)

        if query_result.returncode:
            log.error("Could not determine package contained by rpm at %s", pkg)
            return (None, None)

        # The output from the command is just name|version-release
        name, sep, version = query_result.stdout.strip().partition('|')

        return (name, version)

    @classmethod
    def order_versions(cls, pkg_vers):
        return sorted(pkg_vers, cmp=cls.compare_rpm_versions)

    @classmethod
    def max_version(cls, versions):
        max_ver = None
        for ver in versions:
            if cls.compare_rpm_versions(max_ver, ver) < 0:
                max_ver = ver

        return max_ver

    @classmethod
    def compare_rpm_versions(cls, first_pkg, second_pkg):
        """
        Given two package versions in form VERSION-RELEASE, (-RELEASE optional), compare them
        based on "newness" (where "greater than" equals "newer")
        """

        # Partition the RPM version strings into (VERSION, RELEASE)
        first_fields = first_pkg.split('-', 1) if first_pkg else ()
        second_fields = second_pkg.split('-', 1) if second_pkg else ()

        # Compare VERSION and then RELEASE
        for i in range(2):
            # Build a list of wholly-alpha and wholly-numeric fields; treat non-alphanumeric sequences as separators
            first_chars = re.findall('[a-zA-Z]+|[0-9]+', first_fields[i]) if i < len(first_fields) else []
            second_chars = re.findall('[a-zA-Z]+|[0-9]+', second_fields[i]) if i < len(second_fields) else []

            # Compare position by position
            for j in range(min(len(first_chars), len(second_chars))):
                c1 = first_chars[j]
                c2 = second_chars[j]
                if c1.isdigit():
                    if c2.isdigit():
                        # If both fields are numeric, compare based on int values
                        int_cmp = cmp(int(c1), int(c2))
                        if int_cmp:
                            return int_cmp
                    else:
                        # If one is alpha and one is numeric, then numeric is "greater"
                        return 1
                elif c2.isdigit():
                    # If one is alpha and one is numeric, then numeric is "greater"
                    return -1
                else:
                    # If they're both strings, just compare lexicographically
                    str_cmp = cmp(c1, c2)
                    if str_cmp:
                        return str_cmp

            # If all of the intersecting fields match, the longer string is newer
            len_cmp = cmp(len(first_chars), len(second_chars))
            if len_cmp:
                return len_cmp

        # If both VERSION and RELEASE match for both RPMs, ignoring non-alphanumeric chars, they are equal
        return 0

    def _is_installed(self, pkg):
        pkg_with_version = RpmTool.get_package_version(pkg)

        if not pkg_with_version or not pkg_with_version[0]:
            # If there's an error retrieving the version, assume we have to install it (a failure there will be terminal)
            return True

        pkg_spec = '-'.join(pkg_with_version) if pkg_with_version[1] else pkg_with_version[0]

        # rpm -q will try to find the specific RPM in the local system
        # --quiet will reduce this command to just an exit code
        test_result = ProcessHelper(['rpm', '-q', '--quiet', pkg_spec]).call()

        # if rpm -q returns 0, that means the package exists
        return test_result.returncode == 0





#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
from cfnbootstrap import util
from cfnbootstrap.construction_errors import ToolError
import logging
import os
import re
from cfnbootstrap.packages import requests
import tempfile

log = logging.getLogger("cfn.init")

_msi_supported = True
try:
    import win32com.client
except ImportError:
    _msi_supported = False

class MsiTool(object):

    _remote_pattern = re.compile(r'^(https?|ftp)://.*$', re.I)

    def apply(self, action, auth_config):

        """
        Install a set of MSI packages

        Arguments:
        action -- a dict of package name to path, which is a string

        Exceptions:
        ToolError -- on expected failures (such as a non-zero exit code)
        """

        pkgs_changed = []

        if not action.keys():
            log.debug("No packages installed for MSI")
            return pkgs_changed

        if not _msi_supported:
            raise ToolError("MSI support is only available under Windows")

        pkgs = {}
        tmp_pkgs=[]

        installer_db = Installer()

        try:
            for name, loc in action.iteritems():
                if MsiTool._remote_pattern.match(loc):
                    try:
                        msi_file = self._msi_from_url(loc, auth_config)
                    except IOError, e:
                        raise ToolError("Failed to retrieve %s: %s" % (loc, e.strerror))
                    tmp_pkgs.append(msi_file)
                else:
                    msi_file = loc

                if installer_db.isMsiInstalled(msi_file):
                    log.info("%s is already installed; skipping", name)
                else:
                    pkgs[name] = msi_file

            if not pkgs:
                log.info("All MSI packages already installed")
                return

            for name, pkg in pkgs.iteritems():
                log.debug("Installing %s via MSI", name)
                installer_db.installProduct(pkg)

                log.info("Installed %s successfully", name)
                pkgs_changed.append(pkg)

            return pkgs_changed
        finally:
            for tmp_pkg in tmp_pkgs:
                os.remove(tmp_pkg)

    @util.retry_on_failure()
    def _msi_from_url(self, archive, auth_config):
        tf = tempfile.mkstemp(suffix='.msi', prefix='cfn-init-tmp')

        with os.fdopen(tf[0], 'wb') as temp_dest:
            opts = util.req_opts({'auth': auth_config.get_auth(None)})
            util.EtagCheckedResponse(requests.get(archive, **opts)).write_to(temp_dest)

        return tf[1]

class Installer(object):

    def __init__(self):
        self._installer = win32com.client.gencache.EnsureModule('{000C1092-0000-0000-C000-000000000046}', 1033, 1, 0).Installer()
        self._installer.UILevel = 2 # Hide UI

    def getProperty(self, msi_path, msi_property):
        db = self._installer.OpenDatabase(msi_path, 0)
        view = db.OpenView(r"Select `Value` From Property WHERE `Property` =?")
        try:
            paramRecord = self._installer.CreateRecord(1)
            paramRecord.SetStringData(1, msi_property)
            view.Execute(paramRecord)
            resultRecord = view.Fetch()
            return resultRecord.StringData(1)
        finally:
            view.Close()

    def isMsiInstalled(self, msi_path):
        productCode = self.getProperty(msi_path, "ProductCode")

        try:
            self._installer.ProductInfo(productCode, "VersionString")
            return True
        except Exception:
            return False

    def installProduct(self, msi_path):
        self._installer.InstallProduct(msi_path, "REBOOT=ReallySuppress ALLUSERS=1")


#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
import logging
import security

log = logging.getLogger("cfn.init")

class GroupTool(object):
    """
    Creates OS groups
    """

    def apply(self, action):
        """
        Create groups, returning a list of groups that were created

        Arguments:
        action -- a dict of group name to attributes, where attributes has keys of:
            gid: the gid of the user (a string or int)

        Exceptions:
        ToolError -- on expected failures
        """

        groups_created = []

        if not action:
            log.debug("No groups specified")
            return groups_created

        for name in sorted(action.keys()):
            gid = None
            if "gid" in action[name]:
                gid = str(action[name]["gid"])

            if security.create_group(name, gid):
                groups_created.append(name)

        return groups_created

class UserTool(object):
    """
    Creates OS Users
    """

    def apply(self, action):
        """
        Create users, returning a list of users that were created or modified

        Arguments:
        action -- a dict of user name to attributes, where attributes has keys of:
            groups: A list of group names for this user to be a member of
            homeDir: The home directory for this user
            uid: The uid for this user

        Exceptions:
        ToolError -- on expected failures
        """

        users_modified = []
        if not action:
            log.debug("No users specified")
            return users_modified

        for name in sorted(action.keys()):
            attributes = action[name]

            uid = None if not "uid" in attributes else str(attributes["uid"])

            if security.create_or_modify_user(name, attributes.get("groups", []), attributes.get("homeDir", None), uid):
                users_modified.append(name)

        return users_modified



#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
import threading
from cfnbootstrap import util
from cfnbootstrap.aws_client import AwsQueryError
from cfnbootstrap.cfn_client import CloudFormationClient
from cfnbootstrap.sqs_client import SQSClient
from cfnbootstrap.util import ProcessHelper
from threading import Timer
import cfnbootstrap.json_file_manager as JsonFileManager
import ConfigParser
import calendar
import datetime
import logging
import os
import random
import socket
import subprocess
import tempfile
import time
from util import Credentials

try:
    import simplejson as json
except ImportError:
    import json

log = logging.getLogger("cfn.hup")

def parse_config(config_path):
    main_conf_path = os.path.join(config_path, 'cfn-hup.conf')
    if not os.path.isfile(main_conf_path):
        raise ValueError("Could not find main configuration at %s" % main_conf_path)

    main_config = ConfigParser.SafeConfigParser()
    main_config.read(main_conf_path)

    if not main_config.has_option('main', 'stack'):
        raise ValueError("[main] section must contain stack option")

    stack = main_config.get('main', 'stack')

    if main_config.has_option('main', 'role'):
        credentials = util.RoleBasedCredentials(main_config.get('main', 'role'))
    elif main_config.has_option('main', 'credential-file'):
        try:
            credentials = util.extract_credentials(main_config.get('main', 'credential-file'))
        except IOError, e:
            raise ValueError("Could not retrieve credentials from file:\n\t%s" % e.strerror)
    else:
        credentials = Credentials('', '')

    additional_hooks_path = os.path.join(config_path, 'hooks.d')
    additional_files = []
    if os.path.isdir(additional_hooks_path):
        for hook_file in os.listdir(additional_hooks_path):
            if os.path.isfile(os.path.join(additional_hooks_path, hook_file)):
                additional_files.append(os.path.join(additional_hooks_path, hook_file))

    hooks_config = ConfigParser.SafeConfigParser()
    files_read = hooks_config.read([os.path.join(config_path, 'hooks.conf')] + additional_files)

    if not files_read:
        raise ValueError("No hook configurations found at %s or %s." % (os.path.join(config_path, 'hooks.conf'), additional_hooks_path))

    hooks = []
    cmd_hooks = []

    for section in hooks_config.sections():
        if not hooks_config.has_option(section, 'triggers'):
            logging.error("No triggers specified for hook %s", section)
            continue

        triggers = [s.strip() for s in hooks_config.get(section, 'triggers').split(',')]

        if not hooks_config.has_option(section, 'path'):
            logging.error("No path specified for hook %s", section)
            continue

        if not hooks_config.has_option(section, 'action'):
            logging.error("No action specified for hook %s", section)
            continue

        runas = None
        if hooks_config.has_option(section, 'runas'):
            runas = hooks_config.get(section, 'runas').strip()

        hook = Hook(section,
                    triggers,
                    hooks_config.get(section, 'path').strip(),
                    hooks_config.get(section, 'action'),
                    runas)
        if hook.is_cmd_hook():
            if hooks_config.has_option(section, 'singleton'):
                hook.singleton = util.interpret_boolean(hooks_config.get(section, 'singleton'))
            if hooks_config.has_option(section, 'send_result'):
                hook.send_result = util.interpret_boolean(hooks_config.get(section, 'send_result'))
            cmd_hooks.append(hook)
        else:
            hooks.append(hook)

    if not hooks and not cmd_hooks:
        raise ValueError("No valid hooks found")

    region = 'us-east-1'
    if main_config.has_option('main', 'region'):
        region = main_config.get('main', 'region')

    cfn_url = CloudFormationClient.endpointForRegion(region)

    if main_config.has_option('main', 'url'):
        cfn_url = main_config.get('main', 'url')

    cfn_client = CloudFormationClient(credentials, cfn_url, region)

    if main_config.has_option('main', 'multi-threaded'):
        value = main_config.get('main', 'multi-threaded')
        multi_threaded = util.interpret_boolean(value)
    else:
        multi_threaded = True

    if hooks:
        processor = HookProcessor(hooks, stack, cfn_client)
    else:
        processor = None

    if cmd_hooks:
        sqs_url = SQSClient.endpointForRegion(region)
        if main_config.has_option('main', 'sqs_url'):
            sqs_url = main_config.get('main', 'sqs_url')

        sqs_client = SQSClient(credentials, sqs_url, region=region)

        cmd_processor = CmdProcessor(stack, cmd_hooks, sqs_client,
                                     CloudFormationClient(credentials, cfn_url, region),
                                     multi_threaded)
    else:
        cmd_processor = None

    return (main_config, processor, cmd_processor)

class FatalUpdateError(Exception):

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg

class InFlightStatusError(Exception):

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg

class Hook(object):

    def __init__(self, name, triggers, path, action, runas):
        self._triggers = triggers[:]
        self._path = path
        self._action = action
        self._name = name
        self._runas = runas
        self.singleton = False
        self.send_result = True

    @property
    def triggers(self):
        return self._triggers

    @property
    def path(self):
        return self._path

    @property
    def action(self):
        return self._action

    @property
    def name(self):
        return self._name

    @property
    def runas(self):
        return self._runas

    def is_cmd_hook(self):
        return self._triggers == ['on.command']

class AutoRefreshingCredentialsProvider(object):

    def __init__(self, cfn_client, stack_name, listener_id):
        self._cfn_client = cfn_client
        self._stack_name = stack_name
        self._listener_id = listener_id
        self._creds = None
        self._last_timer = None
        self.listener_expired = False
        self._refresh_lock = threading.Lock()

    def refresh(self):
        with self._refresh_lock:
            log.info("Refreshing listener credentials")
            if self._last_timer:
                self._last_timer.cancel()

            try:
                self._creds = self._cfn_client.get_listener_credentials(self._stack_name, self._listener_id)
                self.listener_expired = False
            except IOError, e:
                if hasattr(e, 'error_code') and 'ListenerExpired' == e.error_code:
                    self.listener_expired = True
                    log.exception("Listener expired")
                else:
                    self.listener_expired = False
                    log.exception("IOError caught while refreshing credentials")
            except Exception:
                self.listener_expired = False
                log.exception("Exception refreshing credentials")

            now = time.time()
            expiration = calendar.timegm(self._creds.expiration.utctimetuple()) if self._creds else now
            remaining = expiration - now

            if remaining > 30 * 60:
                next_refresh = min(2 * 60 * 60, remaining / 2)
            else:
                next_refresh = 60 * random.random()

            log.info("Scheduling next credential refresh in %s seconds", next_refresh)
            t = Timer(next_refresh, self.refresh)
            t.daemon = True
            t.start()
            self._last_timer = t

    def creds_expired(self):
        return self._creds and self._creds.expiration < datetime.datetime.utcnow()

    @property
    def credentials(self):
        for i in range(3):
            if self._creds:
                break
            self.refresh()

        if not self._creds:
            raise ValueError('Could not retrieve listener credentials')

        return self._creds

class CmdProcessor(object):
    """Processes CommandService hooks"""

    def __init__(self, stack_name, hooks, sqs_client, cfn_client, multi_threaded):
        """Takes a list of Hook objects and processes them"""
        self.stack_name = stack_name
        self.hooks = self._hooks_by_path(hooks)
        self.sqs_client = sqs_client
        self.cfn_client = cfn_client
        self.multi_threaded = multi_threaded
        if not self.multi_threaded:
            log.debug("Enabled single threading mode.");

        if util.is_ec2():
            self.listener_id = util.get_instance_id()
        elif not cfn_client.using_instance_identity:
            self.listener_id = socket.getfqdn()
        else:
            raise ValueError("Could not retrieve instance id")

        self._creds_provider = AutoRefreshingCredentialsProvider(self.cfn_client, self.stack_name, self.listener_id)
        self.queue_url = None

        self._create_storage_dir()
        self._runfile = os.path.join(self.storage_dir, 'commands_run.json')

        self._commands_run = self._load_commands_run()

        if not 'by_id' in self._commands_run:
            self._commands_run['by_id'] = {}

        if not 'by_day' in self._commands_run:
            self._commands_run['by_day'] = {}

        self._currently_running = set()

        self._currently_running_lock = threading.RLock()
        self._commands_run_lock = threading.RLock()

    def _load_commands_run(self):
        if os.path.isfile(self._runfile):
            with file(self._runfile, 'r') as f:
                try:
                    return json.load(f)
                except Exception:
                    log.exception("Could not load previously run commands")
                    os.remove(self._runfile)
                    return {}
        else:
            return {}


    def is_registered(self):
        return self.queue_url is not None and not self._creds_provider.listener_expired

    def creds_expired(self):
        return self._creds_provider.creds_expired()

    def register(self):
        try:
            self.queue_url = self.cfn_client.register_listener(self.stack_name, self.listener_id).queue_url
        except AwsQueryError, e:
            if e.error_code == 'ValidationError':
                raise FatalUpdateError('Terminal failure registering listener: %s' % str(e))
            raise
        self._creds_provider.listener_expired = False

    def _create_storage_dir(self):
        if os.name == 'nt':
            self.storage_dir = os.path.expandvars(r'${SystemDrive}\cfn\cfn-hup\data')
        else:
            self.storage_dir = '/var/lib/cfn-hup/data'
        if not os.path.isdir(self.storage_dir):
            log.debug("Creating %s", self.storage_dir)
            try:
                os.makedirs(self.storage_dir, 0700)
            except OSError:
                log.warn("Could not create %s; using temporary directory", self.storage_dir)
                self.storage_dir = tempfile.mkdtemp()

    def _command_completed(self, msg):
        now = datetime.datetime.utcnow()
        cmd_time = now.replace(hour=0, minute=0, second=0, microsecond=0).isoformat()

        with self._commands_run_lock:
            self._commands_run['by_id'][self._get_id(msg)] = True

            if cmd_time not in self._commands_run['by_day']:
                self._commands_run['by_day'][cmd_time] = []

            self._commands_run['by_day'][cmd_time].append(self._get_id(msg))

            try:
                keys_to_delete = []
                for key in self._commands_run['by_day'].iterkeys():
                    if now - datetime.datetime.strptime(key, '%Y-%m-%dT%H:%M:%S') > datetime.timedelta(days=2):
                        for cmd_id in self._commands_run['by_day'][key]:
                            del self._commands_run['by_id'][cmd_id]

                        keys_to_delete.append(key)

                for key in keys_to_delete:
                    del self._commands_run['by_day'][key]

                self._write_commands_run()
            except Exception:
                log.exception('Could not write runfile to %s', self._runfile)

    def _write_commands_run(self):
        with file(self._runfile, 'w') as f:
            json.dump(self._commands_run, f)


    def _delete_message(self, receipt_handle):
        try:
            self.sqs_client.delete_message(self.queue_url, receipt_handle, request_credentials = self._creds_provider.credentials)
        except Exception:
            log.exception("Could not delete message for handle %s", receipt_handle)

    def _get_id(self, msg):
        return msg['DispatcherId'] + '|' + msg['CommandName'] + '|' + msg['InvocationId']

    def _already_run(self, msg):
        with self._commands_run_lock:
            return self._get_id(msg) in self._commands_run['by_id']

    def _parse(self, msg):
        try:
            return json.loads(json.loads(msg.body)['Message'])
        except Exception:
            log.exception("Received invalid message")
            return None

    def process(self):
        if self.queue_url is None:
            raise FatalUpdateError("Cannot process command hooks before registering")

        threads  = []

        try:
            for msg in self.sqs_client.receive_message(self.queue_url, request_credentials = self._creds_provider.credentials, wait_time=20):
                cmd_msg = self._parse(msg)
                if not cmd_msg:
                    log.info("Invalid message, deleting: %s", msg)
                    self._delete_message(msg.receipt_handle)
                    continue

                if self._already_run(cmd_msg):
                    log.info("Already ran %s, deleting", self._get_id(cmd_msg))
                    self._delete_message(msg.receipt_handle)
                    continue

                with self._currently_running_lock:
                    if self._get_id(cmd_msg) not in self._currently_running and not self._already_run(cmd_msg):
                        self._currently_running.add(self._get_id(cmd_msg))

                        if self.multi_threaded:
                            thread = threading.Thread(target=self._process_msg, args=(cmd_msg, msg.receipt_handle))
                            thread.daemon = True
                            threads.append(thread)
                            thread.start()
                        else:
                            self._process_msg(cmd_msg, msg.receipt_handle)

        except FatalUpdateError:
            raise
        except IOError, e:
            if hasattr(e, 'error_code') and 'AWS.SimpleQueueService.NonExistentQueue' == e.error_code:
                self.queue_url = None
            log.exception("IOError caught while processing messages")
        except Exception:
            log.exception("Exception caught while processing messages")

        return threads

    def _process_msg(self, cmd_msg, receipt_handle):
        log.debug("Processing message: %s", cmd_msg)
        delete = False

        try:
            log.debug("Command message: %s", cmd_msg)

            expiration = datetime.datetime.utcfromtimestamp(int(cmd_msg['Expiration']) / 1000)

            if expiration < datetime.datetime.utcnow():
                log.info("Invocation %s of command %s for stack %s expired at %s; skipping",
                            cmd_msg['InvocationId'], cmd_msg['CommandName'], cmd_msg['DispatcherId'],
                            expiration.isoformat())
                delete = True
            else:
                log.info("Received command %s (invocation id: %s)", cmd_msg['CommandName'], cmd_msg['InvocationId'])
                hook_to_run = self.hooks.get(cmd_msg['CommandName'])
                if not hook_to_run or self._run_hook(hook_to_run, cmd_msg):
                    self._command_completed(cmd_msg)
                    delete = True
        except (ValueError, KeyError):
            log.exception("Invalid message received; deleting it")
            delete = True
        except Exception:
            log.exception("Unexpected exception running command")
        finally:
            if delete:
                self._delete_message(receipt_handle)
            with self._currently_running_lock:
                self._currently_running.remove(self._get_id(cmd_msg))

    def _run_hook(self, hook, cmd_msg):
        if hook.singleton:
            log.info("Hook %s is configured to run as a singleton", hook.name)
            leader = self.cfn_client.elect_command_leader(self.stack_name,
                                                          cmd_msg['CommandName'],
                                                          cmd_msg['InvocationId'],
                                                          self.listener_id)
            if leader == self.listener_id:
                log.info("This listener is the leader.  Continuing with action")
            else:
                log.info("This listener is not the leader; %s is the leader.", leader)
                return True

        action_env = self._get_environment(cmd_msg)
        result_queue = cmd_msg['ResultQueue']

        log.info("Running action for %s", hook.name)
        log.debug("Action environment: %s", action_env)

        action = hook.action
        if hook.runas:
            if os.name == 'posix':
                action = ['su', hook.runas, '-c', action]
            else:
                log.warn('runas is not supported on this operating system')

        result = ProcessHelper(action, stderr=subprocess.PIPE, env=action_env).call()

        log.debug("Action for %s output: %s", hook.name, result.stdout if result.stdout else '<None>')

        if not hook.send_result:
            return True

        result_msg = { 'DispatcherId' : cmd_msg['DispatcherId'],
                       'InvocationId' : cmd_msg['InvocationId'],
                       'CommandName' : cmd_msg['CommandName'],
                       'Status' : "FAILURE" if result.returncode else "SUCCESS",
                       'ListenerId' : self.listener_id }

        if result.returncode:
            result_stderr = result.stderr.rstrip()
            log.warn("Action for %s exited with %s, returning FAILURE", hook.name, result.returncode)
            result_msg['Message'] = result_stderr if len(result_stderr) <= 1024 else result_stderr[0:500] + '...' + result_stderr[-500:]
        else:
            result_stdout = result.stdout.rstrip()
            if len(result_stdout) > 1024:
                log.error("stdout for %s was greater than 1024 in length, which is not allowed", hook.name)
                result_msg['Status'] = 'FAILURE'
                result_msg['Message'] = 'Result data was longer than 1024 bytes. Started with: ' + result_stdout[0:100]
            else:
                log.info("Action for %s succeeded, returning SUCCESS", hook.name)
                result_msg['Data'] = result_stdout

        try:
            self.sqs_client.send_message(result_queue, json.dumps(result_msg), request_credentials=self._creds_provider.credentials)
        except Exception:
            log.exception('Error sending result; will leave message in queue')
            return False

        return True

    def _hooks_by_path(self, hooks):
        ret_hooks = {}
        for hook in hooks:
            if hook.path in ret_hooks:
                raise FatalUpdateError("Multiple hooks for the same command (%s)" % hook.path)
            ret_hooks[hook.path] = hook
        return ret_hooks

    def _get_environment(self, cmd_msg):
        action_env = dict(os.environ)
        if 'Data' in cmd_msg:
            action_env['CMD_DATA'] = cmd_msg['Data']
        action_env['INVOCATION_ID'] = cmd_msg['InvocationId']
        action_env['DISPATCHER_ID'] = cmd_msg['DispatcherId']
        action_env['CMD_NAME'] = cmd_msg['CommandName']
        action_env['STACK_NAME'] = self.stack_name
        action_env['LISTENER_ID'] = self.listener_id
        action_env['RESULT_QUEUE'] = cmd_msg['ResultQueue']
        if 'EventHandle' in cmd_msg:
            action_env['EVENT_HANDLE'] = cmd_msg['EventHandle']
        creds = self._creds_provider.credentials
        action_env['RESULT_ACCESS_KEY'] = creds.access_key
        action_env['RESULT_SECRET_KEY'] = creds.secret_key
        action_env['RESULT_SESSION_TOKEN'] = creds.security_token
        return action_env

class HookProcessor(object):
    """Processes update hooks"""

    def __init__(self, hooks, stack_name, client):
        """Takes a list of Hook objects and processes them"""
        self.hooks = hooks
        if os.name == 'nt':
            self.dir = os.path.expandvars(r'${SystemDrive}\cfn\cfn-hup\data')
        else:
            self.dir = '/var/lib/cfn-hup/data'
        if not os.path.isdir(self.dir):
            log.debug("Creating %s", self.dir)
            try:
                os.makedirs(self.dir, 0700)
            except OSError:
                log.warn("Could not create %s; using temporary directory", self.dir)
                self.dir = tempfile.mkdtemp()
        JsonFileManager.create(self.dir, 'metadata_db.json')
        self.client = client
        self.stack_name = stack_name
        self._jsonConverter = JsonFileManager.Converter([])

    def process(self):
        metadata = JsonFileManager.read(self.dir, 'metadata_db.json')
        self._resource_cache = {}
        for hook in self.hooks:
            try:
                self._process_hook(hook, metadata)
            except FatalUpdateError:
                raise
            except Exception:
                log.exception("Exception caught while running hook %s", hook.name)

    def _process_hook(self, hook, metadata):
        try:
            new_data = self._retrieve_path_data(hook.path)
        except InFlightStatusError:
            return

        old_data = self._jsonConverter.deserialize(metadata.get(hook.name + "|" + hook.path, None))

        if 'post.add' in hook.triggers and not old_data and new_data:
            log.info("Previous state not found; action for %s will be run", hook.name)
        elif 'post.remove' in hook.triggers and old_data and not new_data:
            log.info('Path %s was removed; action for %s will be run', hook.path, hook.name)
        elif 'post.update' in hook.triggers and old_data and new_data and old_data != new_data:
            log.info("Data has changed from previous state; action for %s will be run", hook.name)
        else:
            log.debug("No change in path %s for hook %s", hook.path, hook.name)
            metadata[hook.name + "|" + hook.path] = self._jsonConverter.serialize(new_data)
            JsonFileManager.write(self.dir, 'metadata_db.json', metadata)
            return

        log.info("Running action for %s", hook.name)
        action_env = dict(os.environ)
        env_key = self._retrieve_env_key(hook.path)
        if old_data:
            action_env['CFN_OLD_%s' % env_key] = self._as_string(old_data)
        if new_data:
            action_env['CFN_NEW_%s' % env_key] = self._as_string(new_data)

        action = hook.action
        if hook.runas:
            action = ['su', hook.runas, '-c', action]

        result = ProcessHelper(action, env=action_env).call()

        if result.returncode:
            log.warn("Action for %s exited with %s; will retry on next iteration", hook.name, result.returncode)
        else:
            metadata[hook.name + '|' + hook.path] = self._jsonConverter.serialize(new_data)
            JsonFileManager.write(self.dir, 'metadata_db.json', metadata)
        log.debug("Action for %s output: %s", hook.name, result.stdout if result.stdout else '<None>')

    def _as_string(self, obj):
        if isinstance(obj, basestring):
            return obj
        elif isinstance(obj, datetime.datetime):
            return obj.isoformat()
        return json.dumps(obj)

    def _retrieve_env_key(self, path):
        """Given a hook path, return the key to append to environment variables for old/new data"""
        parts = path.split('.', 3)

        if len(parts) < 3:
            return 'LAST_UPDATED'
        elif parts[2].lower() == 'metadata':
            return 'METADATA'
        elif parts[2].lower() == 'physicalresourceid':
            return 'PHYSICAL_RESOURCE_ID'

    def _retrieve_path_data(self, path):
        parts = path.split('.', 3)
        if len(parts) < 2:
            raise FatalUpdateError("Unsupported path: paths must be in the form Resources.<LogicalResourceId>(.Metadata|PhysicalResourceId)(.<optional Metadata subkey>). Input: %s" % path)

        if parts[0].lower() != 'resources':
            raise FatalUpdateError('Unsupported path: only changes to Resources are supported (path: %s)' % path)

        if len(parts) == 2:
            resourcePart = None
        elif parts[2].lower() not in ['metadata', 'physicalresourceid']:
            raise FatalUpdateError("Unsupported path: only Metadata or PhysicalResourceId can be specified after LogicalResourceId (path: %s)" % path)
        else:
            resourcePart = parts[2].lower()

        logical_id = parts[1]
        subpath = ('' if len(parts) < 4 else parts[3])

        if logical_id not in self._resource_cache:
            self._resource_cache[logical_id] = self.client.describe_stack_resource(logical_id, self.stack_name)

        resource = self._resource_cache[logical_id]
        status = resource.resourceStatus

        if status and status.endswith('_IN_PROGRESS'):
            log.debug("Skipping resource %s in %s as it is in status %s", logical_id, self.stack_name, status)
            raise InFlightStatusError('%s in %s is in status %s' % (logical_id, self.stack_name, status))

        if resourcePart == 'metadata':
            if not resource.metadata:
                log.warn("No metadata for %s in %s", logical_id, self.stack_name)
                return None

            return util.extract_value(resource.metadata, subpath)
        elif 'DELETE_COMPLETE' == status:
            return None
        elif resourcePart == 'physicalresourceid':
            return resource.physicalResourceId
        else:
            return resource.lastUpdated



#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================

# Wrapper for the CFN logging operations in order to avoid string encoding problems
# Still using logging.Logger methods, but provided with UTF-8 encoded arguments
# Takes the (unique) logger returned by logging.getLogger()
class CfnLogger:
    def __init__(self, logger):
        self._log = logger
    
    def debug(self, msg, *args, **kwargs):
        self._log.debug(to_unicode(msg), *to_unicode_list(args), **to_unicode_kwargs_dictionary(kwargs))

    def info(self, msg, *args, **kwargs):
        self._log.info(to_unicode(msg), *to_unicode_list(args), **to_unicode_kwargs_dictionary(kwargs))

    def warning(self, msg, *args, **kwargs):
        self._log.warning(to_unicode(msg), *to_unicode_list(args), **to_unicode_kwargs_dictionary(kwargs))

    def error(self, msg, *args, **kwargs):
        self._log.error(to_unicode(msg), *to_unicode_list(args), **to_unicode_kwargs_dictionary(kwargs))

    def critical(self, msg, *args, **kwargs):
        self._log.critical(to_unicode(msg), *to_unicode_list(args), **to_unicode_kwargs_dictionary(kwargs))

    def log(self, msg, *args, **kwargs):
        self._log.log(to_unicode(msg), *to_unicode_list(args), **to_unicode_kwargs_dictionary(kwargs))

    def exception(self, msg, *args, **kwargs):
        self._log.exception(to_unicode(msg), *to_unicode_list(args), **to_unicode_kwargs_dictionary(kwargs))

# If str, convert it to Unicode - otherwise untouched
def to_unicode(x):
    if isinstance(x, str):
        return x.decode('utf-8')
    return x

def to_unicode_list(a_list):
    return [to_unicode(x) for x in a_list]

# Convert it to Unicode only the key (of the dictionary) that equals the filter - the rest of the keys stay untouched
def to_unicode_filter_dict(a_key, a_dictionary, a_filter):
    if a_key == a_filter:
        return to_unicode(a_dictionary[a_key])
    return a_dictionary[a_key]

# kwargs['extra'] might also have a non-Unicode encoding
def to_unicode_kwargs_dictionary(a_dictionary):
    retval = {}
    for x in a_dictionary.keys():
        retval[x] = to_unicode_filter_dict(x, a_dictionary, 'extra')
    return retval




"""
Copyright (c) Donald Stufft, pip, and individual contributors

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from __future__ import absolute_import

import sys


class VendorAlias(object):

    def __init__(self, package_names):
        self._package_names = package_names
        self._vendor_name = __name__
        self._vendor_pkg = self._vendor_name + "."
        self._vendor_pkgs = [
            self._vendor_pkg + name for name in self._package_names
        ]

    def find_module(self, fullname, path=None):
        if fullname.startswith(self._vendor_pkg):
            return self

    def load_module(self, name):
        # Ensure that this only works for the vendored name
        if not name.startswith(self._vendor_pkg):
            raise ImportError(
                "Cannot import %s, must be a subpackage of '%s'." % (
                    name, self._vendor_name,
                )
            )

        if not (name == self._vendor_name or
                any(name.startswith(pkg) for pkg in self._vendor_pkgs)):
            raise ImportError(
                "Cannot import %s, must be one of %s." % (
                    name, self._vendor_pkgs
                )
            )

        # Check to see if we already have this item in sys.modules, if we do
        # then simply return that.
        if name in sys.modules:
            return sys.modules[name]

        # Check to see if we can import the vendor name
        try:
            # We do this dance here because we want to try and import this
            # module without hitting a recursion error because of a bunch of
            # VendorAlias instances on sys.meta_path
            real_meta_path = sys.meta_path[:]
            try:
                sys.meta_path = [
                    m for m in sys.meta_path
                    if not isinstance(m, VendorAlias)
                ]
                __import__(name)
                module = sys.modules[name]
            finally:
                # Re-add any additions to sys.meta_path that were made while
                # during the import we just did, otherwise things like
                # requests.packages.urllib3.poolmanager will fail.
                for m in sys.meta_path:
                    if m not in real_meta_path:
                        real_meta_path.append(m)

                # Restore sys.meta_path with any new items.
                sys.meta_path = real_meta_path
        except ImportError:
            # We can't import the vendor name, so we'll try to import the
            # "real" name.
            real_name = name[len(self._vendor_pkg):]
            try:
                __import__(real_name)
                module = sys.modules[real_name]
            except ImportError:
                raise ImportError("No module named '%s'" % (name,))

        # If we've gotten here we've found the module we're looking for, either
        # as part of our vendored package, or as the real name, so we'll add
        # it to sys.modules as the vendored name so that we don't have to do
        # the lookup again.
        sys.modules[name] = module

        # Finally, return the loaded module
        return module


sys.meta_path.append(VendorAlias(["urllib3", "chardet"]))



#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================

from cfnbootstrap.util import ProcessHelper
import logging
from cfn_logging import CfnLogger
import os
from cfnbootstrap.construction_errors import ToolError
import subprocess

log = CfnLogger(logging.getLogger("cfn.init"))

class AptTool(object):
    """
    Installs packages via APT

    """

    def apply(self, action, auth_config=None):
        """
        Install a set of packages via APT, returning the packages actually installed or updated.

        Arguments:
        action -- a dict of package name to version; version can be empty, a single string or a list of strings

        Exceptions:
        ToolError -- on expected failures (such as a non-zero exit code)
        """

        pkgs_changed = []

        if not action:
            log.debug("No packages specified for APT")
            return pkgs_changed

        cache_result = ProcessHelper(['apt-cache', '-q', 'gencaches']).call()

        if cache_result.returncode:
            log.error("APT gencache failed. Output: %s", cache_result.stdout)
            raise ToolError("Could not create apt cache", cache_result.returncode)

        pkg_specs = []

        for pkg_name in action:
            if action[pkg_name]:
                if isinstance(action[pkg_name], basestring):
                    pkg_keys = ['%s=%s' % (pkg_name, action[pkg_name])]
                else:
                    pkg_keys = ['%s=%s' % (pkg_name, ver) if ver else pkg_name for ver in action[pkg_name]]
            else:
                pkg_keys = [pkg_name]

            pkgs_filtered = [pkg_key for pkg_key in pkg_keys if self._pkg_filter(pkg_key, pkg_name)]
            if pkgs_filtered:
                pkg_specs.extend(pkgs_filtered)
                pkgs_changed.append(pkg_name)

        if not pkg_specs:
            log.info("All APT packages were already installed")
            return []

        log.info("Attempting to install %s via APT", pkg_specs)

        env = dict(os.environ)
        env['DEBIAN_FRONTEND'] = 'noninteractive'

        result = ProcessHelper(['apt-get', '-q', '-y', 'install'] + pkg_specs, env=env).call()

        if result.returncode:
            log.error("apt-get failed. Output: %s", result.stdout)
            raise ToolError("Could not successfully install APT packages", result.returncode)

        log.info("APT installed %s", pkgs_changed)
        log.debug("APT output: %s", result.stdout)

        return pkgs_changed

    def _pkg_filter(self, pkg, pkg_name):
        if self._pkg_installed(pkg, pkg_name):
            log.debug("%s will not be installed as it is already present", pkg)
            return False
        elif not self._pkg_available(pkg):
            log.error("%s is not available to be installed", pkg)
            raise ToolError("APT does not have %s available for installation" % pkg)
        else:
            return True

    def _pkg_available(self, pkg):
        result = ProcessHelper(['apt-cache', '-q', 'show', pkg]).call()

        return result.returncode == 0

    def _pkg_installed(self, pkg, pkg_name):
        """
        Test if a package is installed (exact version match if version is specified), returning a boolean.

        Arguments:
        pkg -- the full package specification (including version if specified) in pkg=version format
        pkg_name -- the name of the package
        """

        result = ProcessHelper(['dpkg-query', '-f', '${Status}|${Package}=${Version}', '-W', pkg_name], stderr=subprocess.PIPE).call()

        if result.returncode or not result.stdout:
            return False

        status,divider,spec = result.stdout.strip().partition('|')

        if status.rpartition(" ")[2] != 'installed':
            return False

        return spec.startswith(pkg)



#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
"""
SQS client-related classes

Classes:
SQSClient - an HTTP client that makes API calls against SQS
Message  - a message from an SQS queue

"""
from cfnbootstrap import aws_client
from cfnbootstrap.aws_client import V4Signer
from cfnbootstrap.util import retry_on_failure, timeout
from cfnbootstrap import endpoint_tool
from xml.etree import ElementTree
import StringIO
import logging
import re

log = logging.getLogger("cfn.client")

class SQSClient(aws_client.Client):
    """
    Makes API calls against SQS

    Notes:
    - Public methods of this class have a 1-to-1 equivalence to published SQS APIs.
    - Calls are retried internally when appropriate; callers should not retry.

    Attributes:
    _apiVersion - the SQS API version
    _xmlns      - the XML namespace for the SQS API version in use

    """

    _apiVersion = "2012-11-05"
    _xmlns='http://queue.amazonaws.com/doc/%s/' % _apiVersion

    def __init__(self, credentials, url=None, region='us-east-1', proxyinfo=None):
        if not url:
            endpoint = SQSClient.endpointForRegion(region)
        else:
            endpoint = self._fix_endpoint(url)

        if not region:
            region = SQSClient.regionForEndpoint(endpoint)

        if not region:
            raise ValueError('Region is required for AWS V4 Signatures')

        signer = V4Signer(region, 'sqs')

        super(SQSClient, self).__init__(credentials, False, endpoint, signer=signer, xmlns='http://queue.amazonaws.com/doc/%s/' % SQSClient._apiVersion, proxyinfo=proxyinfo)
        log.debug("SQS client initialized with endpoint %s", endpoint)

    # SQS SSL certificates have CNs based on sqs.<region>.amazonaws.com
    # Python2.6 will fail to verify the hostname of the certificate
    # Due to http://bugs.python.org/issue13034 only being fixed in 2.7 and 3.2
    def _fix_endpoint(self, unparsed_url):
        # Some endpoints with older certificates still use <region>.queue.amazonaws.com instead
        m = re.match(endpoint_tool.SQS_PATTERN, unparsed_url) or re.match(endpoint_tool.SQS_LEGACY_PATTERN, unparsed_url)
        if m:
            for endpoint in endpoint_tool.get_endpoints_for_service("AmazonSQS"):
                if endpoint.region == m.group(1):
                    fixed_hostname = endpoint.hostname
                    if fixed_hostname.endswith('.cn') and m.group(2).startswith('.cn'):
                        fixed_hostname = fixed_hostname[:-3]
                    return 'https://%s%s' % (fixed_hostname, m.group(2))
            return 'https://' + endpoint_tool.SQS_FORMAT % (m.group(1), m.group(2))
        return unparsed_url

    @classmethod
    def endpointForRegion(cls, region):
        for endpoint in endpoint_tool.get_endpoints_for_service("AmazonSQS"):
            if endpoint.region == region:
                return 'https://%s' % endpoint.hostname
        return 'https://' + endpoint_tool.SQS_FORMAT % (region, '.cn' if region.startswith('cn') else '')

    @classmethod
    def regionForEndpoint(cls, parsed_url):
        for endpoint in endpoint_tool.get_endpoints_for_service("AmazonSQS"):
            if parsed_url == 'https://%s' % endpoint.hostname:
                return endpoint.region
        m = re.match(endpoint_tool.SQS_PATTERN, parsed_url) or re.match(endpoint_tool.SQS_LEGACY_PATTERN, parsed_url)
        if m:
            return m.group(1)
        return None

    @retry_on_failure(http_error_extractor=aws_client.Client._get_xml_extractor(_xmlns))
    @timeout(60)
    def receive_message(self, queue_url, attributes=None, max_messages=1, visibility_timeout=None,
                              request_credentials=None, wait_time=None):
        """
        Calls ReceiveMessage and returns a list of Message objects

        Throws an IOError on failure.
        """
        if not attributes: attributes = ['All']
        queue_url = self._fix_endpoint(queue_url)
        log.debug("Receiving messages for queue %s", queue_url)

        params = { "Action" : "ReceiveMessage", "Version" : SQSClient._apiVersion, "MaxNumberOfMessages" : str(max_messages) }
        for i in range(len(attributes)):
            params['AttributeName.%s' % (i + 1)]=attributes[i]
        if visibility_timeout:
            params['VisibilityTimeout'] = str(visibility_timeout)
        if wait_time:
            params['WaitTimeSeconds'] = str(wait_time)

        response_content = self._call(params, queue_url, request_credentials,
                                      timeout=wait_time + 3 if wait_time else None).content
        return Message._parse_list(StringIO.StringIO(response_content), self._xmlns)

    @retry_on_failure(max_tries=25, http_error_extractor=aws_client.Client._get_xml_extractor(_xmlns))
    @timeout()
    def send_message(self, queue_url, message_body, delay_seconds=None, request_credentials=None):
        """
        Calls SendMessage and returns a tuple of (MessageId, MD5OfMessageBody)

        Throws an IOError on failure.
        """
        queue_url = self._fix_endpoint(queue_url)
        log.debug("Sending %s to queue %s", message_body, queue_url)

        params = { "Action" : "SendMessage", "Version" : SQSClient._apiVersion, "MessageBody" : message_body}

        if delay_seconds:
            params["DelaySeconds"] = delay_seconds

        root = ElementTree.ElementTree(file=StringIO.StringIO(self._call(params, queue_url, request_credentials, verb='POST').content))
        message_id = root.findtext('{%s}SendMessageResult/{%s}MessageId' % (self._xmlns, self._xmlns))
        md5_of_body = root.findtext('{%s}SendMessageResult/{%s}MD5OfMessageBody' % (self._xmlns, self._xmlns))

        return (message_id, md5_of_body)

    @retry_on_failure(http_error_extractor=aws_client.Client._get_xml_extractor(_xmlns))
    @timeout()
    def delete_message(self, queue_url, receipt_handle, request_credentials=None):
        """
        Calls DeleteMessage on a specified receipt handle

        Throws an IOError on failure.
        """
        queue_url = self._fix_endpoint(queue_url)
        log.debug("Deleting %s from queue %s", receipt_handle, queue_url)

        params = { "Action" : "DeleteMessage", "Version" : SQSClient._apiVersion, "ReceiptHandle" : receipt_handle}

        self._call(params, queue_url, request_credentials)

class Message(object):
    """A message off of an SQS queue"""

    @classmethod
    def _parse_list(cls, data, xmlns):
        if not data:
            return []
        root = ElementTree.ElementTree(file=data)
        msgs = root.findall('{%s}ReceiveMessageResult/{%s}Message' % (xmlns, xmlns))
        return [cls._from_elem(elem, xmlns) for elem in msgs]

    @classmethod
    def _from_elem(cls, elem, xmlns):
        attribs = {}
        for attrib in elem.findall('{%s}Attribute' % xmlns):
            attribs[attrib.findtext('{%s}Name' % xmlns)] = attrib.findtext('{%s}Value' % xmlns)

        return Message(elem.findtext('{%s}MessageId' % xmlns),
                       elem.findtext('{%s}ReceiptHandle' % xmlns),
                       elem.findtext('{%s}MD5OfBody' % xmlns),
                       elem.findtext('{%s}Body' % xmlns),
                       attribs)

    def __init__(self, msg_id, handle, md5, body, attribs):
        self._message_id = msg_id
        self._receipt_handle = handle
        self._md5_of_body = md5
        self._body = body
        self._attributes = dict(attribs)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return 'Message[MessageId:%s, ReceiptHandle:%s, MD5OfBody:%s, Body:%s, Attributes:%s]' % (self.message_id,
                                                                                                  self.receipt_handle,
                                                                                                  self.md5_of_body,
                                                                                                  self.body,
                                                                                                  self.attributes)

    @property
    def message_id(self):
        return self._message_id

    @property
    def receipt_handle(self):
        return self._receipt_handle

    @property
    def md5_of_body(self):
        return self._md5_of_body

    @property
    def body(self):
        return self._body

    @property
    def attributes(self):
        return self._attributes



#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
from resources import documents
import urlparse
import re

S3_PATTERN = r'https://s3([\.-][a-z0-9-]+)\.amazonaws\.com(.*)$'
S3_WITH_BUCKET_PATTERN = r'https://([a-z0-9-\.]+)[\.-]s3([\.-][a-z0-9-]+)\.amazonaws\.com(.*)$'
SQS_PATTERN = r'https://sqs\.([\w-]+)\.amazonaws\.com(.*)$'
SQS_LEGACY_PATTERN = r'https://([\w-]+)\.queue.amazonaws\.com(.*)$'

S3_FORMAT = 's3%s.amazonaws.com%s'
SQS_FORMAT = 'sqs.%s.amazonaws.com%s'
SQS_LEGACY_FORMAT = '%s.queue.amazonaws.com%s'

_endpoint_data = documents.get_endpoint_data()

def get_endpoints_for_service(service):
    """Returns unique service endpoints for special regions."""

    return [Endpoint.from_data(e) for e in _endpoint_data["Services"][service]["Endpoints"]]

def is_service_url(service, unparsed_url):
    return get_endpoint_for_url(service, unparsed_url) is not None

def get_endpoint_for_url(service, unparsed_url):
    for endpoint in get_endpoints_for_service(service):
        if endpoint.matches_url(unparsed_url):
            return endpoint
    if service == "AmazonS3":
        m = re.match(S3_WITH_BUCKET_PATTERN, unparsed_url)
        if m:
            return Endpoint.from_region(S3_FORMAT, m.group(2))
        else:
            m = re.match(S3_PATTERN, unparsed_url)
            if m:
                return Endpoint.from_region(S3_FORMAT, m.group(1))
    elif service == "AmazonSQS":
        m = re.match(SQS_PATTERN, unparsed_url)
        if m:
            return Endpoint.from_region(SQS_FORMAT, m.group(1))
        else:
            m = re.match(SQS_LEGACY_PATTERN, unparsed_url)
            if m:
                return Endpoint.from_region(SQS_LEGACY_FORMAT, m.group(1))

    return None

class Endpoint(object):
    """
    Represents an AWS service endpoint
    """

    @classmethod
    def from_data(cls, endpoint_data):
        is_default = False
        if "Default" in endpoint_data:
            is_default = bool(endpoint_data["Default"])
        return cls(endpoint_data["Region"], endpoint_data["Hostname"], is_default)

    @classmethod
    def from_region(cls, endpoint_format, region):
        if region[:1] not in ['.', '-']:
            region = '.' + region
        return cls(region[1:], endpoint_format % (region, '.cn' if region.startswith('cn') else ''))

    def __init__(self, region, hostname, is_default=False):
        if region is None:
            raise ValueError("region is required")
        if hostname is None:
            raise ValueError("hostname is required")

        self.region = region
        self.hostname = hostname
        self.is_default = is_default

    def matches_url(self, unparsed_url):
        return urlparse.urlparse(unparsed_url).netloc.lower().endswith(self.hostname)

    def get_subdomain_prefix(self, unparsed_url):
        netloc = urlparse.urlparse(unparsed_url).netloc.lower()
        if not netloc.endswith(self.hostname):
            return None
        return netloc.rpartition(self.hostname)[0]



#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================

import logging
import sys

_scheduler_supported = True
try:
    import win32com.client
    from pythoncom import com_error
except ImportError:
    _scheduler_supported = False

log = logging.getLogger("cfn.init")

def set_reboot_trigger():
    if not _scheduler_supported:
        log.debug("Not setting a reboot trigger as scheduling support is not available")
        return

    scheduler = win32com.client.gencache.EnsureModule('{E34CB9F1-C7F7-424C-BE29-027DCC09363A}', 0, 1, 0).TaskScheduler()
    scheduler.Connect()

    log.debug("Creating Scheduled Task for cfn-init resume")

    root_folder = scheduler.GetFolder("\\")
    task_definition = scheduler.NewTask(0) # This must always be 0

    registration_info = task_definition.RegistrationInfo
    registration_info.Description = "This will resume cfn-init after the system has booted"
    registration_info.Author = "Amazon Web Services"

    task_definition.Settings.StartWhenAvailable = True
    task_definition.Settings.RunOnlyIfNetworkAvailable = True
    task_definition.Settings.RestartCount = 3
    task_definition.Settings.RestartInterval = "PT1M"

    trigger = task_definition.Triggers.Create(win32com.client.constants.TASK_TRIGGER_BOOT)
    trigger = win32com.client.CastTo(trigger, "IBootTrigger")
    trigger.Id = "CfnInitBootTrigger"
    trigger.Delay = "PT30S"

    action = task_definition.Actions.Create(win32com.client.constants.TASK_ACTION_EXEC)
    action = win32com.client.CastTo(action, "IExecAction")
    action.Path = sys.executable
    action.Arguments = "-v --resume"

    task_definition.Principal.UserId="NT AUTHORITY\SYSTEM"
    task_definition.Principal.RunLevel=win32com.client.constants.TASK_RUNLEVEL_HIGHEST

    try:
        root_folder.RegisterTaskDefinition("cfn-init Resume Trigger",
                                           task_definition,
                                           win32com.client.constants.TASK_CREATE,
                                           None,
                                           None,
                                           win32com.client.constants.TASK_LOGON_SERVICE_ACCOUNT)
    except com_error:
        log.debug("Scheduled task already exists; not updating")
    else:
        log.debug("Scheduled Task created")


def clear_reboot_trigger():
    if not _scheduler_supported:
        log.debug("Not clearing reboot trigger as scheduling support is not available")
        return
    
    scheduler = win32com.client.gencache.EnsureModule('{E34CB9F1-C7F7-424C-BE29-027DCC09363A}', 0, 1, 0).TaskScheduler()
    scheduler.Connect()

    log.debug("Deleting Scheduled Task for cfn-init resume")

    root_folder = scheduler.GetFolder("\\")
    try:
        root_folder.DeleteTask("cfn-init Resume Trigger", 0)
        log.debug("Scheduled Task deleted")
    except com_error:
        log.debug("Scheduled Task did not exist")


#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
"""
CloudFormation client-related classes

Classes:
CloudFormationClient - an HTTP client that makes API calls against CloudFormation
StackResourceDetail  - detailed information about a StackResource

"""
from cfnbootstrap import aws_client, util
from cfnbootstrap.aws_client import CFNSigner, V4Signer
from cfnbootstrap.util import retry_on_failure, timeout
import datetime
import logging
import re
from util import Credentials

try:
    import simplejson as json
except ImportError:
    import json

log = logging.getLogger("cfn.client")

class CloudFormationClient(aws_client.Client):
    """
    Makes API calls against CloudFormation

    Notes:
    - Public methods of this class have a 1-to-1 equivalence to published CloudFormation APIs.
    - Calls are retried internally when appropriate; callers should not retry.

    Attributes:
    _apiVersion - the CloudFormation API version
    _endpoints  - CloudFormation service endpoints differing from the https://cloudformation.<region>.amazonaws.com format

    """

    _apiVersion = "2010-05-15"
    _endpoints = { "cn-north-1": "https://cloudformation.cn-north-1.amazonaws.com.cn",
                   "cn-northwest-1": "https://cloudformation.cn-northwest-1.amazonaws.com.cn"}

    def __init__(self, credentials, url=None, region='us-east-1', proxyinfo=None):

        if not url:
            endpoint = CloudFormationClient.endpointForRegion(region)
        else:
            endpoint = url

        self.using_instance_identity = (not credentials or not credentials.access_key) and util.is_ec2()

        if not self.using_instance_identity and (not credentials or not credentials.access_key or not credentials.secret_key):
            raise ValueError('In order to sign requests, 169.254.169.254 must be accessible or valid credentials must '
                             'be set. Please ensure your proxy environment variables allow access to 169.254.169.254 '
                             '(NO_PROXY) or that your credentials have a valid access key and secret key.')

        if not self.using_instance_identity:
            if not region:
                region = CloudFormationClient.regionForEndpoint(endpoint)

            if not region:
                raise ValueError('Region is required for AWS V4 Signatures')

        signer = CFNSigner() if self.using_instance_identity else V4Signer(region, 'cloudformation')

        super(CloudFormationClient, self).__init__(credentials, True, endpoint, signer, proxyinfo=proxyinfo)

        log.debug("CloudFormation client initialized with endpoint %s", endpoint)

    @classmethod
    def endpointForRegion(cls, region):
        if region in CloudFormationClient._endpoints:
            return CloudFormationClient._endpoints[region]
        return 'https://cloudformation.%s.amazonaws.com' % region

    @classmethod
    def regionForEndpoint(cls, endpoint):
        match = re.match(r'https://cloudformation.([\w\d-]+).amazonaws.com', endpoint)
        if match:
            return match.group(1)
        log.warn("Non-standard CloudFormation endpoint: %s", endpoint)
        return None

    @retry_on_failure(http_error_extractor=aws_client.Client._extract_json_message)
    @timeout()
    def describe_stack_resource(self, logicalResourceId, stackName, request_credentials=None):
        """
        Calls DescribeStackResource and returns a StackResourceDetail object.

        Throws an IOError on failure.
        """
        log.debug("Describing resource %s in stack %s", logicalResourceId, stackName)

        return StackResourceDetail(self._call({"Action" : "DescribeStackResource",
                                               "LogicalResourceId" : logicalResourceId,
                                               "ContentType" : "JSON",
                                               "StackName": stackName,
                                               "Version": CloudFormationClient._apiVersion },
                                               request_credentials=request_credentials))

    @retry_on_failure(http_error_extractor=aws_client.Client._extract_json_message)
    @timeout()
    def signal_resource(self, logicalResourceId, stackName, uniqueId, status="SUCCESS", request_credentials=None):
        """
        Calls SignalResource.

        Throws an IOError on failure.
        """
        log.debug("Signaling resource %s in stack %s with unique ID %s and status %s", logicalResourceId,
                                                                                       stackName,
                                                                                       uniqueId,
                                                                                       status)
        self._call({"Action": "SignalResource",
                    "LogicalResourceId": logicalResourceId,
                    "StackName": stackName,
                    "UniqueId": uniqueId,
                    "Status": status,
                    "ContentType": "JSON",
                    "Version": CloudFormationClient._apiVersion },
                    request_credentials=request_credentials)

    @retry_on_failure(http_error_extractor=aws_client.Client._extract_json_message)
    @timeout()
    def register_listener(self, stack_name, listener_id=None, request_credentials=None):
        """
        Calls RegisterListener and returns a Listener object

        Throws an IOError on failure.
        """
        log.debug("Registering listener %s for stack %s", listener_id, stack_name)

        params = {"Action" : "RegisterListener",
                  "StackName" : stack_name,
                  "ContentType" : "JSON"}

        if not self.using_instance_identity:
            params["ListenerId"] = listener_id

        return Listener(self._call(params, request_credentials = request_credentials))

    @retry_on_failure(http_error_extractor=aws_client.Client._extract_json_message)
    @timeout()
    def elect_command_leader(self, stack_name, command_name, invocation_id, listener_id=None, request_credentials=None):
        """
        Calls ElectCommandLeader and returns the listener id of the leader

        Throws an IOError on failure.
        """
        log.debug("Attempting to elect '%s' as leader for stack: %s, command: %s, invocation: %s",
                  listener_id, stack_name, command_name, invocation_id)

        params = {"Action" : "ElectCommandLeader",
                  "CommandName" : command_name,
                  "InvocationId" : invocation_id,
                  "StackName" : stack_name,
                  "ContentType" : "JSON"}

        if not self.using_instance_identity:
            params["ListenerId"] = listener_id

        result_data = self._call(params, request_credentials = request_credentials).json()

        return result_data['ElectCommandLeaderResponse']['ElectCommandLeaderResult']['ListenerId']

    @retry_on_failure(http_error_extractor=aws_client.Client._extract_json_message)
    @timeout()
    def get_listener_credentials(self, stack_name, listener_id=None, request_credentials=None):
        """
        Calls GetListenerCredentials and returns a Credentials object

        Throws an IOError on failure.
        """
        log.debug("Get listener credentials for listener %s in stack %s", listener_id, stack_name)

        params = {"Action" : "GetListenerCredentials",
                  "StackName" : stack_name,
                  "ContentType" : "JSON"}

        if not self.using_instance_identity:
            params["ListenerId"] = listener_id

        resp = self._call(params, request_credentials = request_credentials)
        body = resp.json()['GetListenerCredentialsResponse']['GetListenerCredentialsResult']['Credentials']
        return Credentials(body['AccessKeyId'],
            body['SecretAccessKey'],
            body['SessionToken'],
            datetime.datetime.utcfromtimestamp(body['Expiration']))


class Listener(object):
    """Result of RegisterListener"""

    def __init__(self, resp):
        result = resp.json()['RegisterListenerResponse']['RegisterListenerResult']
        self._queue_url = result['QueueUrl']

    @property
    def queue_url(self):
        return self._queue_url

class StackResourceDetail(object):
    """Detailed information about a stack resource"""

    def __init__(self, resp):
        detail = resp.json()['DescribeStackResourceResponse']['DescribeStackResourceResult']['StackResourceDetail']

        self._description = detail.get('Description')
        self._lastUpdated = datetime.datetime.utcfromtimestamp(detail['LastUpdatedTimestamp'])
        self._logicalResourceId = detail['LogicalResourceId']

        _rawMetadata = detail.get('Metadata')
        self._metadata = json.loads(_rawMetadata) if _rawMetadata else None

        self._physicalResourceId = detail.get('PhysicalResourceId')
        self._resourceType = detail['ResourceType']
        self._resourceStatus = detail['ResourceStatus']
        self._resourceStatusReason = detail.get('ResourceStatusReason')
        self._stackId = detail.get('StackId')
        self._stackName = detail.get('StackName')

    @property
    def logicalResourceId(self):
        """The resource's logical resource ID"""
        return self._logicalResourceId

    @property
    def description(self):
        """The resource's description"""
        return self._description

    @property
    def lastUpdated(self):
        """The timestamp of this resource's last status change as a datetime object"""
        return self._lastUpdated

    @property
    def metadata(self):
        """The resource's metadata as python object (not as a JSON string)"""
        return self._metadata

    @property
    def physicalResourceId(self):
        """The resource's physical resource ID"""
        return self._physicalResourceId

    @property
    def resourceType(self):
        """The resource's type"""
        return self._resourceType

    @property
    def resourceStatus(self):
        """The resource's status"""
        return self._resourceStatus

    @property
    def resourceStatusReason(self):
        """The reason for this resource's status"""
        return self._resourceStatusReason

    @property
    def stackId(self):
        """The ID of the stack this resource belongs to"""
        return self._stackId

    @property
    def stackName(self):
        """The name of the stack this resource belongs to"""
        return self._stackName



#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
from cfnbootstrap import update_hooks
import cfnbootstrap
import logging
import os
import time

try:
    import servicemanager
    import win32event
    import win32service
    import win32serviceutil
except ImportError:
    logging.warn("Win32 cfn-hup service requires pywin32")

class HupService(win32serviceutil.ServiceFramework):
    _svc_name_ = 'cfn-hup'
    _svc_display_name_="CloudFormation cfn-hup"
    def __init__(self, args):
        win32serviceutil.ServiceFramework.__init__(self, args)
        self.hWaitStop = win32event.CreateEvent(None, 0, 0, None)
        
    def SvcStop(self):
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        servicemanager.LogMsg(servicemanager.EVENTLOG_INFORMATION_TYPE,
                              servicemanager.PYS_SERVICE_STOPPED,
                              (self._svc_name_, ''))
        win32event.SetEvent(self.hWaitStop)
        
    def SvcDoRun(self):
        try:
            main_config, processor, cmd_processor = update_hooks.parse_config(os.path.expandvars('${SystemDrive}\cfn'))
        except ValueError, e:
            servicemanager.LogMsg(servicemanager.EVENTLOG_ERROR_TYPE,
                                  servicemanager.PYS_SERVICE_STOPPING,
                                  (self._svc_name_, ': %s' % str(e)))
            return
        
        verbose = main_config.has_option('main', 'verbose') and main_config.getboolean('main', 'verbose')
        cfnbootstrap.configureLogging("DEBUG" if verbose else "INFO", filename='cfn-hup.log')
        log = logging.getLogger("cfn.hup")
        
        if main_config.has_option('main', 'interval'):
            interval = main_config.getint('main', 'interval')
            if interval < 1:
                log.error("Invalid interval (must be 1 minute or greater): %s", interval)
                interval = 15
        else:
            interval = 15
                
        interval = interval * 60
        
        servicemanager.LogMsg(servicemanager.EVENTLOG_INFORMATION_TYPE,
                              servicemanager.PYS_SERVICE_STARTED,
                              (self._svc_name_, ''))
        
        last_normal_hup = 0
        delay = 1 * 1000
        while True:
            try:
                if processor and time.time() - last_normal_hup > interval:
                    processor.process()
                    last_normal_hup = time.time()
                
                if cmd_processor:
                    if not cmd_processor.is_registered():
                        cmd_processor.register()
                    
                    if cmd_processor.creds_expired():
                        delay = 20 * 1000
                        log.error("Expired credentials found; skipping process")
                    else:
                        delay = 1 * 1000
                        cmd_processor.process()
            except update_hooks.FatalUpdateError:
                log.exception("Fatal exception caught; stopping cfn-hup")
                break
            except Exception:
                log.exception("Unhandled exception")
            
            if win32event.WAIT_OBJECT_0 == win32event.WaitForSingleObject(self.hWaitStop, delay):
                log.info("Received shutdown event; stopping cfn-hup")
                break
            
    


#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================

class BuildError(Exception):
    """
    Base exception for errors raised while building
    """

    pass

class NoSuchConfigSetError(BuildError):
    """
    Exception signifying no config error with specified name exists
    """
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg

class NoSuchConfigurationError(BuildError):
    """
    Exception signifying no config error with specified name exists
    """
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg

class CircularConfigSetDependencyError(BuildError):
    """
    Exception signifying circular dependency in configSets
    """
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg

class ToolError(BuildError):
    """
    Exception raised by Tools when they cannot successfully change reality

    Attributes:
    msg - a human-readable error message
    code - an error code, if applicable
    """

    def __init__(self, msg, code=None):
        self.msg = msg
        self.code = code

    def __str__(self):
        if (self.code):
            return '%s (return code %s)' % (self.msg, self.code)
        else:
            return self.msg



#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
from cfnbootstrap import util
from cfnbootstrap.construction_errors import ToolError
from cfnbootstrap.util import retry_on_failure
from tarfile import TarError
from zipfile import BadZipfile
import logging
import os.path
from os import chmod 
import re
from cfnbootstrap.packages import requests
import shutil
import tarfile
import tempfile
import zipfile

log = logging.getLogger("cfn.init")

class SourcesTool(object):
    """
    Explodes sources (archives) into locations on disk

    """

    _remote_pattern = re.compile(r'^(https?|ftp)://.*$', re.I)
    _github_pattern = re.compile(r'^https?://github.com/.*?/(zipball|tarball)/.*$')

    def apply(self, action, auth_config):
        """
        Extract archives to their corresponding destination directories, returning directories which were updated.

        Arguments:
        action -- a dict of directory to archive location, which can be either a path or URL
        auth_config -- an AuthenticationConfig object for managing authenticated downloads

        Exceptions:
        ToolError -- on expected failures
        """

        dirs_changed = []

        if not action:
            log.debug("No sources specified")
            return dirs_changed

        for (path, archive) in sorted(action.iteritems(), key=lambda pair: pair[0]):

            if SourcesTool._remote_pattern.match(archive):
                try:
                    archive_file = self._archive_from_url(archive, auth_config)
                except IOError, e:
                    raise ToolError("Failed to retrieve %s: %s" % (archive, e.strerror))
            else:
                if not os.path.isfile(archive):
                    raise ToolError("%s does not exist" % archive)
                archive_file = file(archive, 'rb')

            if TarWrapper.is_compatible(archive_file):
                log.debug("Treating %s as a tarball", archive)
                archive_wrapper = TarWrapper(archive_file)
            elif ZipWrapper.is_compatible(archive_file):
                log.debug("Treating %s as a zip archive", archive)
                archive_wrapper = ZipWrapper(archive_file)
            else:
                raise ToolError("Unsupported source file (not zip or tarball): %s" % archive)

            log.debug("Checking to ensure that all archive members fall under path %s" % path)
            self._check_all_members_in_path(path, archive_wrapper)

            if SourcesTool._github_pattern.match(archive.lower()):
                log.debug("Attempting to magically strip GitHub parent directory from archive")
                archive_wrapper=self._perform_github_magic(archive_wrapper);

            log.debug("Expanding %s into %s", archive, path)
            archive_wrapper.extract_all(path)
            dirs_changed.append(path)

        return dirs_changed

    def _perform_github_magic(self, archive):
        """
        The tarballs that GitHub autogenerates via their HTTP API put the contents
        of the tree into a top-level directory that has no value or predictability.
        This essentially "strips" that top level directory -- unfortunately, python has no
        equivalent of tar's --strip-components.  So we unarchive it and then rearchive
        a subtree.
        """
        tempdir = tempfile.mkdtemp()
        try:
            archive.extract_all(tempdir)
            tempmembers = os.listdir(tempdir)
            if len(tempmembers) != 1:
                log.debug("GitHub magic is not possible; archive does not contain exactly one directory")
                return archive
            else:
                temparchive = tempfile.TemporaryFile()
                tf = tarfile.TarFile(fileobj=temparchive, mode='w')
                parent = os.path.join(tempdir, tempmembers[0])
                log.debug("Creating temporary tar from %s", parent)
                for member in os.listdir(parent):
                    tf.add(os.path.join(parent, member), arcname=member)
                tf.close()
                temparchive.seek(0, 0)
                return TarWrapper(temparchive)
        finally:
            shutil.rmtree(tempdir, True)

    def _check_all_members_in_path(self, path, archive):
        """
        This does a best-effort test to make sure absolute paths
        or ../../../../ nonsense in archives makes files "escape"
        their destination
        """

        normalized_parent = os.path.normcase(os.path.abspath(path))
        for member in archive.files():
            if os.path.isabs(member):
                prefix = os.path.commonprefix([os.path.normcase(os.path.normpath(member)), normalized_parent])
            else:
                prefix = os.path.commonprefix([os.path.normcase(os.path.normpath(os.path.join(normalized_parent, member))), normalized_parent])

            if prefix != normalized_parent:
                raise ToolError("%s is not a sub-path of %s" % (member, path))

    @retry_on_failure()
    def _archive_from_url(self, archive, auth_config):
        tf = tempfile.TemporaryFile()
        opts = util.req_opts({'auth': auth_config.get_auth(None)})
        util.EtagCheckedResponse(requests.get(archive, **opts)).write_to(tf)
        tf.flush()
        tf.seek(0, os.SEEK_SET)
        return tf

class ZipWrapper(object):

    def __init__(self, f):
        self.file = zipfile.ZipFile(f, mode='r')
        # change encoding of all filenames to UTF-8 so that Unicode chars aren't treated as ASCII
        for info in self.file.infolist():
            if type(info.filename) != unicode:
                info.filename = unicode(info.filename, 'utf8')

    @classmethod
    def is_compatible(cls, f):
        try:
            z = zipfile.ZipFile(f, mode='r')
            z.close()
            f.seek(0, 0)
            return True
        except BadZipfile:
            return False

    def files(self):
        return (info.filename for info in self.file.infolist())

    def extract_all(self, dest):
        """
        The zipfile module doesn't preserve file permissions when extracting, so
        each file needs to have its mode changed via chmod after extracting it.
        The file's original permission bits are stored in the external attributes of 
        the file's ZipInfo object, and we retrieve them by shifting right 16 bits.
        """
        
        
        
        for info in self.file.infolist():
            ext_attr = info.external_attr
            mode = ext_attr >> 16
            target_path = self.file.extract(info, dest)
            # ignoring chmod for Windows
            if os.name != 'nt':
                chmod(target_path, mode)

class TarWrapper(object):

    def __init__(self, f):
        self.file = tarfile.open(fileobj = f, mode='r:*')
        # change encoding of all filenames to UTF-8 so that Unicode chars aren't treated as ASCII
        for member in self.file.getmembers():
            member.name = unicode(member.name, 'utf8')

    @classmethod
    def is_compatible(cls, f):
        try:
            t = tarfile.open(fileobj = f, mode='r:*')
            t.close()
            f.seek(0, 0)
            return True
        except TarError:
            return False

    def files(self):
        return self.file.getnames()

    def extract_all(self, dest):
        self.file.extractall(dest)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
certs.py
~~~~~~~~

This module returns the preferred default CA certificate bundle.

If you are packaging Requests, e.g., for a Linux distribution or a managed
environment, you can change the definition of where() to return a separately
packaged CA bundle.
"""

import os
import sys
import imp


def where():
    """Return the preferred certificate bundle."""

    if os.name == 'nt':
        override_path = os.path.expandvars(r'${SystemDrive}\cfn\ca-override.pem')
    else:
        override_path = '/etc/cfn/ca-override.pem'

    override_path = os.getenv('CA_OVERRIDE', override_path)

    if os.path.isfile(override_path):
        return override_path

    # Modification for compilation by py2exe:
    if (hasattr(sys, "frozen") or  # new py2exe
            hasattr(sys, "importers")  # old py2exe
            or imp.is_frozen("__main__")):  # tools/freeze
        return os.path.join(os.path.dirname(sys.executable), 'cacert.pem')
    # end py2exe

    # vendored bundle inside Requests
    return os.path.join(os.path.dirname(__file__), 'cacert.pem')

if __name__ == '__main__':
    print(where())



#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
import logging.config
import os.path
import sys
import StringIO
import stat
import os

class NullHandler(logging.Handler):
    def emit(self, record):
        pass

_config ="""[loggers]
keys=root,cfninit,cfnclient,cfnhup,wire,cmd
[handlers]
keys=%(all_handlers)s,null
[formatters]
keys=amzn,amzn_cmd
[logger_root]
level=NOTSET
handlers=%(root_handler)s
[logger_cfninit]
level=NOTSET
handlers=%(root_handler)s
qualname=cfn.init
propagate=0
[logger_wire]
level=NOTSET
handlers=%(wire_handler)s
qualname=wire
propagate=0
[logger_cfnhup]
level=NOTSET
handlers=%(root_handler)s
qualname=cfn.hup
propagate=0
[logger_cfnclient]
level=NOTSET
handlers=%(root_handler)s
qualname=cfn.client
propagate=0
[logger_cmd]
level=NOTSET
handlers=%(cmd_handler)s
qualname=cfn.init.cmd
propagate=0
[handler_default]
class=handlers.RotatingFileHandler
level=%(conf_level)s
formatter=amzn
args=('%(conf_file)s', 'a', 5242880, 5, 'UTF-8')
[handler_wire]
class=handlers.RotatingFileHandler
level=DEBUG
formatter=amzn
args=('%(wire_file)s', 'a', 5242880, 5, 'UTF-8')
[handler_null]
class=cfnbootstrap.NullHandler
args=()
[handler_cmd]
class=handlers.RotatingFileHandler
level=DEBUG
formatter=amzn_cmd
args=('%(cmd_file)s', 'a', 5242880, 5, 'UTF-8')
[handler_tostderr]
class=StreamHandler
level=%(conf_level)s
formatter=amzn
args=(sys.stderr,)
[formatter_amzn]
format=%(asctime)s [%(levelname)s] %(message)s
datefmt=
class=logging.Formatter
[formatter_amzn_cmd]
format=%(asctime)s P%(process)d [%(levelname)s] %(message)s
datefmt=
class=logging.Formatter
"""

def resetGroupWorldWritable(fullname):
    current_file_mode = stat.S_IMODE(os.stat(fullname).st_mode)
    # reset group and world writable attribute
    current_file_mode = current_file_mode & 0755
    os.chmod(fullname, current_file_mode)

def _getLogFile(log_dir, filename):
    if log_dir:
        return os.path.join(log_dir, filename)
    if os.name == 'nt':
        logdir = os.path.expandvars(r'${SystemDrive}\cfn\log')
        if not os.path.exists(logdir):
            os.makedirs(logdir)
        return logdir + os.path.sep + filename
    return '/var/log/%s' % filename


def configureLogging(level='INFO', quiet=False, filename='cfn-init.log', log_dir=None, wire_log=True, cmd_log=True):

    output_file = _getLogFile(log_dir, filename)

    config = {'conf_level': level,
              'root_handler' : 'default',
              'conf_file': output_file}
    all_handlers = ["default"]

    if wire_log:
      wire_file =_getLogFile(log_dir,'cfn-wire.log')
      config['wire_file'] = wire_file
      config['wire_handler'] = 'wire' 
      all_handlers.append("wire")
    else:
      config['wire_handler'] = 'null' 

    if cmd_log:
      cmd_file =_getLogFile(log_dir,'cfn-init-cmd.log') 
      config['cmd_file'] = cmd_file
      config['cmd_handler'] = 'cmd' 
      all_handlers.append("cmd")
    else:
      config['cmd_handler'] = 'null' 

    config['all_handlers'] = ','.join(all_handlers)

    try:
        logging.config.fileConfig(StringIO.StringIO(_config), config)
        # make sure any generated log files are not world or group writable
        resetGroupWorldWritable(output_file)
        if wire_log:
            resetGroupWorldWritable(wire_file)
        if cmd_log:
            resetGroupWorldWritable(cmd_file)
    except IOError:
        config['all_handlers'] = 'tostderr'
        config['root_handler'] = 'tostderr'
        config['wire_handler'] = 'null'
        config['cmd_handler']  = 'null'
        if not quiet:
            print >> sys.stderr, "Could not open %s for logging.  Using stderr instead." % output_file
        logging.config.fileConfig(StringIO.StringIO(_config), config)



#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
import os
import logging

def set_owner_and_group(filename, owner_name, group_name):
    logging.warn("Unsupported OS for setting owner/group: %s" % os.name)

def create_or_modify_user(user_name, groups=[], homedir=None, uid=None):
    logging.warn("Unsupported OS for user operations: %s", os.name)

def create_group(group_name, gid=None):
    logging.warn("Unsupported OS for group operations: %s", os.name)

if os.name == "posix":
    from posix_security import *



#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
from cfnbootstrap.util import ProcessHelper
import logging
from cfnbootstrap.construction_errors import ToolError
import re

log = logging.getLogger("cfn.init")

class PythonTool(object):
    """
    Installs packages via easy_install

    """

    def apply(self, action, auth_config=None):
        """
        Install a set of packages via easy_install, returning the packages actually installed or updated.

        Arguments:
        action -- a dict of package name to version; version can be empty, a single string or a list of strings

        Exceptions:
        ToolError -- on expected failures (such as a non-zero exit code)
        """

        pkgs_changed = []

        if not action.keys():
            log.debug("No packages specified for python")
            return pkgs_changed

        pkgs = []

        for pkg in action:
            if not action[pkg] or isinstance(action[pkg], basestring):
                pkgs.append(PythonTool._pkg_spec(pkg, action[pkg]))
            else:
                pkgs.extend(PythonTool._pkg_spec(pkg, ver) for ver in action[pkg])

            pkgs_changed.append(pkg)

        log.info("Attempting to install %s via easy_install", pkgs)

        result = ProcessHelper(['easy_install'] + pkgs).call()

        if result.returncode:
            log.error("easy_install failed. Output: %s", result.stdout)
            raise ToolError("Could not successfully install python packages", result.returncode)
        else:
            log.info("easy_install installed %s", pkgs)
            log.debug("easy_install output: %s", result.stdout)

        return pkgs_changed

    _url_pattern = re.compile(r'^https?://.*$')

    @classmethod
    def _pkg_spec(cls, pkg, ver):
        if not ver:
            return pkg
        if cls._url_pattern.match(ver.lower()):
            return ver
        return '%s==%s' % (pkg, ver)

class GemTool(object):
    """
    Installs packages via rubygems

    """

    def apply(self, action, auth_config=None):
        """
        Install a set of packages via rubygems, returning the packages actually installed or updated.

        Arguments:
        action -- a dict of package name to version; version can be empty, a single string or a list of strings

        Exceptions:
        ToolError -- on expected failures (such as a non-zero exit code)
        """

        pkgs_changed = []

        if not action.keys():
            log.debug("No gems specified")
            return pkgs_changed

        for pkg in action:
            installed = False
            if not action[pkg]:
                installed = self._install_gem(pkg)
            else:
                if isinstance(action[pkg], basestring):
                    installed = self._install_gem(pkg, action[pkg])
                else:
                    for ver in action[pkg]:
                        if self._install_gem(pkg, ver):
                            installed = True

            if installed:
                pkgs_changed.append(pkg)

        return pkgs_changed

    def _gem_is_installed(self, pkg, ver=None):
        """"
        Check to see if a package at version ver is installed.
        If ver is not specified, just check for the package.
        """
        log.debug("Checking to see if %s-%s is already installed", pkg, ver)

        queryCmd = ['gem', 'query', '-i', '-n', '^%s$' % pkg]

        if ver:
            queryCmd.extend(['-v', '%s' % ver])

        result = ProcessHelper(queryCmd).call()

        if result.returncode:
            return False
        else:
            return True

    def _install_gem(self, pkg, ver=None):
        """Install a gem if the version is not already installed; return True if installed, False if skipped."""
        if self._gem_is_installed(pkg, ver):
            log.info("%s-%s is already installed, skipping.", pkg, ver)
            return False
        else:
            log.info("Installing %s version %s via gem", pkg, ver)

            install_command = ['gem', 'install', '-b', '--no-ri', '--no-rdoc', pkg];

            if ver:
                install_command.extend(['-v', '= %s' % ver])

            result = ProcessHelper(install_command).call()

            if result.returncode:
                log.error("Gem failed. Output: %s", result.stdout)
                raise ToolError("Failed to install gem: %s-%s" % (pkg, ver), result.returncode)
            else:
                log.info("Gem installed: %s-%s", pkg, ver)
                log.debug("Gem output: %s", result.stdout)
                return True


#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
import copy
import imp
import logging
import os
import sys

# Imports define which python packages are added to the windows installer when we build it
# Importing pkg_resources causes import errors since six.py is a dependency it cannot resolve
# as the rest of setuptools is missing, so we don't import it in windows exe's as it isn't used anyway
def _is_from_exe():
    return (hasattr(sys, "frozen") or  # new py2exe
           hasattr(sys, "importers")  # old py2exe
           or imp.is_frozen("__main__"))

if not _is_from_exe():
    import pkg_resources

log = logging.getLogger(__name__)

try:
    import simplejson as json
except ImportError:
    import json

if os.name == 'nt':
    endpoint_override_path = os.path.expandvars(r"${SystemDrive}\cfn\endpoints-override.json")
else:
    endpoint_override_path = "/etc/cfn/endpoints-override.json"

#Overwrite path with environment variable if it exists (either UNIX or Windows)
endpoint_override_path = os.getenv("ENDPOINTS_OVERRIDE", endpoint_override_path)


try:
    #Modification for compilation by py2exe:
    if _is_from_exe():
        #Load endpoints.json from .exe directory, i.e. C:\Program Files\Amazon\cfn-bootstrap
        log.debug('Loading py2exe endpoints.json file')
        with open(os.path.join(os.path.dirname(sys.executable), 'endpoints.json'), 'r') as f:
            _endpoint_data = json.load(f)

    # end py2exe
    else:
        _endpoint_data = json.load(pkg_resources.resource_stream("cfnbootstrap.resources.documents", "endpoints.json"))

except ValueError:
    log.exception("Failed to load endpoints.json")
    raise

try:
    #Apply endpoint_override, if present
    if os.path.isfile(endpoint_override_path):
        log.debug('Loading existing endpoints override file %s' % (endpoint_override_path,))
        with open(endpoint_override_path, 'r') as f:
            _endpoint_data = json.load(f)
except ValueError:
    log.exception("Failed to load endpoints override file: %s" % (endpoint_override_path,))
    raise


def get_endpoint_data():
    return copy.deepcopy(_endpoint_data)


#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
from cfnbootstrap.construction_errors import ToolError
from cfnbootstrap.util import ProcessHelper, LoggingProcessHelper, interpret_boolean
import logging
import os.path
import subprocess
try:
    import simplejson as json
except ImportError:
    import json

log = logging.getLogger("cfn.init")

class CommandTool(object):
    """
    Executes arbitrary commands
    """

    def apply(self, action):
        """
        Execute a set of commands, returning a list of commands that were executed.

        Arguments:
        action -- a dict of command to attributes, where attributes has keys of:
            command: the command to run (a string or list)
            cwd: working directory (a string)
            env: a dictionary of environment variables
            test: a commmand to run; if it returns zero, the command will run
            ignoreErrors: if true, ignore errors
            waitAfterCompletion: # of seconds to wait after completion (or "forever")
            defaults: a command to run; the stdout will be used to provide defaults

        Exceptions:
        ToolError -- on expected failures
        """

        commands_run = []

        if not action:
            log.debug("No commands specified")
            return commands_run

        for name in sorted(action.keys()):
            log.debug(u"Running command %s", name)

            attributes = action[name]

            if "defaults" in attributes:
                log.debug(u"Generating defaults for command %s", name)
                defaultsResult = ProcessHelper(attributes['defaults'], stderr=subprocess.PIPE).call()
                log.debug(u"Defaults script for %s output: %s", name, defaultsResult.stdout.decode('utf-8'))
                if defaultsResult.returncode:
                    log.error(u"Defaults script failed for %s: %s", name, defaultsResult.stderr.decode('utf-8'))
                    raise ToolError(u"Defaults script for command %s failed" % name)

                default_attrs = attributes
                default_env = default_attrs.get("env",{})
                attributes = json.loads(defaultsResult.stdout)
                user_env = attributes.get("env",{})
                user_env.update(default_env)
                attributes.update(default_attrs)
                attributes["env"] = user_env

            if not "command" in attributes:
                log.error(u"No command specified for %s", name)
                raise ToolError(u"%s does not specify the 'command' attribute, which is required" % name)

            cwd = os.path.expanduser(attributes["cwd"]) if "cwd" in attributes else None
            env = attributes.get("env", None)

            if "test" in attributes:
                log.debug(u"Running test for command %s", name)
                test = attributes["test"]
                testResult = LoggingProcessHelper(test, name=u'Test for Command %s' % name, env=env, cwd=cwd).call()
                log.debug(u"Test command output: %s", testResult.stdout.decode('utf-8'))
                if testResult.returncode:
                    log.info(u"Test failed with code %s", testResult.returncode)
                    continue
                else:
                    log.debug(u"Test for command %s passed", name)
            else:
                log.debug(u"No test for command %s", name)

            cmd_to_run = attributes["command"]
            if "runas" in attributes:
                if os.name == 'nt':
                    raise ToolError(u'Command %s specified "runas", which is not supported on Windows' % name)

                if isinstance(cmd_to_run, basestring):
                    cmd_to_run = u'su %s -c %s' % (attributes['runas'], cmd_to_run)
                else:
                    cmd_to_run = ['su', attributes['runas'], '-c'] + cmd_to_run

            commandResult = LoggingProcessHelper(cmd_to_run, name=u'Command %s' % name, env=env, cwd=cwd).call()

            if commandResult.returncode:
                log.error(u"Command %s (%s) failed", name, attributes["command"])
                log.debug(u"Command %s output: %s", name, commandResult.stdout.decode('utf-8'))
                if interpret_boolean(attributes.get("ignoreErrors")):
                    log.info("ignoreErrors set to true, continuing build")
                    commands_run.append(name)
                else:
                    raise ToolError(u"Command %s failed" % name)
            else:
                log.info(u"Command %s succeeded", name)
                log.debug(u"Command %s output: %s", name, commandResult.stdout.decode('utf-8'))
                commands_run.append(name)

        return commands_run

    @classmethod
    def get_wait(cls, cmd_options):
        wait = cmd_options.get('waitAfterCompletion', 60 if os.name == 'nt' else 0)
        if isinstance(wait, basestring) and 'forever' == wait.lower():
            return -1
        return int(wait)



#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================

# Json format related operations are wrapped in this module to make moking of those operations easier

import os
import copy
import collections
import stat
import time
import datetime
import decimal

try:
    import simplejson as json
except ImportError:
    import json

def setSafeMode(directory, filename):
    fullname = os.path.join(directory, filename)
    currentFileMode = stat.S_IMODE(os.stat(fullname).st_mode)
    currentFileMode = currentFileMode & 0600
    os.chmod(fullname, currentFileMode)

def create(directory, filename):
    if not os.path.isfile(os.path.join(directory, filename)):
        open(os.path.join(directory, filename), 'w').close()
    setSafeMode(directory, filename)
    
def read(directory, filename):
    try:
        with open(os.path.join(directory, filename), 'r') as fp:
            metadata = json.load(fp)
    except:
        metadata = {}
    return metadata

def write(directory, filename, metadata):
    with open(os.path.join(directory, filename), 'w') as fp:
        json.dump(metadata, fp)
    setSafeMode(directory, filename)

# JSON doesn't serialize deques and user defined types. Defining here our own conversion.
class Converter():
    def __init__(self, userDefinedTypes):
        # Markers for the manually serialized types
        # collection.deque is treated specially
        # datetime.datetime is also treated specially
        # User defined types are all treated uniformly
        self._userDefinedTypes = userDefinedTypes
        self._deque_marker = '_deque: special_serialization: not_regular_json_'
        self._datetime_marker = '_datetime: special_serialization: not_regular_json_'
        self._markers = {}
        for t in self._userDefinedTypes:
            self._markers[t] = '_' + t.__name__ + ': special_serialization: not_regular_json_'
        # The format used to serialize datetime
        self._datetime_format = '"%Y-%m-%dT%H:%M:%S.%f"'

    def serialize(self, data):
        raw = copy.deepcopy(data)
        if isinstance(raw, collections.deque):
            retval = [self.serialize(v) for v in raw]
            return {self._deque_marker : retval}
        elif isinstance(raw, datetime.datetime):
            retval = self._datetime2str(raw)
            return {self._datetime_marker : retval}
        else:
            for t in self._userDefinedTypes:
                if isinstance(raw, t):
                    return raw.serialize(self._markers[t])
            return raw

    def deserialize(self, data):
        serialized = copy.deepcopy(data)
        if not isinstance(serialized, dict):
            return serialized
        if self._deque_marker in serialized:
            serialized = serialized[self._deque_marker]
            return collections.deque([self.deserialize(v) for v in serialized])
        elif self._datetime_marker in serialized:
            serialized = serialized[self._datetime_marker]
            return self._str2datetime(serialized)
        else:
            for t in self._userDefinedTypes:
                if self._markers[t] in serialized:
                    serialized = serialized[self._markers[t]]
                    return t.from_json(serialized)
            return serialized

    def _datetime2str(self, d):
        return unicode(d.strftime(self._datetime_format))

    def _str2datetime(self, t):
        return datetime.datetime.strptime(t, self._datetime_format)

